# -*- coding: utf-8 -*-
# ---
# jupyter:
#   jupytext:
#     cell_metadata_filter: -all
#     custom_cell_magics: kql
#     text_representation:
#       extension: .py
#       format_name: percent
#       format_version: '1.3'
#       jupytext_version: 1.16.4
#   kernelspec:
#     display_name: gpjax
#     language: python
#     name: python3
# ---

# %% [markdown]
# # Introduction to Bayesian Optimisation
#
# In this guide we introduce the Bayesian Optimisation (BO) paradigm for
# optimising black-box functions. We'll assume an understanding of Gaussian processes
# (GPs), so if you're not familiar with them, check out our [GP introduction notebook](https://docs.jaxgaussianprocesses.com/examples/intro_to_gps/).

# %%
from typing import (
    List,
    Tuple,
)

import jax

# Enable Float64 for more stable matrix inversions.
from jax import config
import jax.numpy as jnp
import jax.random as jr
from jaxopt import ScipyBoundedMinimize
from jaxtyping import (
    Float,
    Int,
    install_import_hook,
)
import matplotlib as mpl
from matplotlib import cm
import matplotlib.pyplot as plt
import optax as ox
import tensorflow_probability.substrates.jax as tfp

from examples.utils import use_mpl_style
from gpjax.parameters import Static
from gpjax.typing import (
    Array,
    FunctionalSample,
    ScalarFloat,
)

config.update("jax_enable_x64", True)


with install_import_hook("gpjax", "beartype.beartype"):
    import gpjax as gpx


# set the default style for plotting
use_mpl_style()

key = jr.key(42)

cols = mpl.rcParams["axes.prop_cycle"].by_key()["color"]


# %% [markdown]
# ## Some Motivating Examples
#
# Countless problems in the physical world involve optimising functions for which the
# explicit functional form is unknown, but which can be expensively queried throughout
# their domain. For example, within the domain of science the task of designing new
# molecules with optimised properties ([Griffiths and Lobato,
# 2020](https://pubs.rsc.org/en/content/articlehtml/2019/sc/c9sc04026a)) is incredibly
# useful. Here, the domain being optimised over is the space of possible molecules, with
# the objective function depending on the property being optimised, for instance within
# drug-design this may be the efficacy of the drug. The function from molecules to
# efficacy is unknown, but can be queried by synthesising a molecule and running an
# experiment to measure its efficacy. This is clearly an expensive procedure!
#
# Within the domain of machine learning, the task of optimising neural network
# architectures is another example of such a problem (commonly referred to as [Neural
# Architecture Search (NAS)](https://en.wikipedia.org/wiki/Neural_architecture_search)).
# Here, the domain is the space of possible neural network architectures, and the
# objective function is a metric such as the accuracy of the trained model. Again, the
# function from neural network architectures to accuracy is unknown, but can be queried by
# training a model with a given architecture and evaluating its accuracy. This is also an
# expensive procedure, as training models can be incredibly time consuming and
# computationally demanding.
#
# Finally, these problems are ubiquitous within the field of climate science, with
# ([Hellan et al., 2023](https://arxiv.org/abs/2306.04343)) providing several excellent
# examples. One such example is the task of deciding where to place wind turbines in a
# wind farm in order to maximise the energy generated. Here, the domain is the space of
# possible locations for the wind turbines, and the objective function is the energy
# generated by the wind farm. The function from locations to energy generated is unknown,
# but could be queried by running a simulation of the wind farm with the turbines placed
# at a given set of locations. Running such simulations can be expensive, particularly if
# they are high-fidelity.
#
# At the heart of all these problems is the task of optimising a function for which we
# don't have the explicit functional form, but which we can (expensively) query at any
# point in its domain. Bayesian optimisation provides a principled framework for solving
# such problems.

# %% [markdown]
# ## What is Bayesian Optimisation?
#
# Bayesian optimisation (BO) ([Močkus, 1974](https://link.springer.com/chapter/10.1007/3-540-07165-2_55)) provides a principled
# method for making decisions under uncertainty. The aim of BO is to find the global
# minimum of a *black-box* objective function, $\min_{\mathbf{x} \in X}
# f(\mathbf{x})$. The function $f$ is said to be a *black-box* function because its
# explicit functional form is unknown. However, it is assumed that one is able to
# ascertain information about the function by evaluating it at points in its domain,
# $X$. However, these evaluations are assumed to be *expensive*, as seen in the
# motivating examples. Therefore, the goal of BO is to minimise $f$ with as few
# evaluations of the black-box function as possible.
#
# As such, BO can be thought of as *sequential decision-making* problem. At each iteration
# one must choose which point (or batch of points) in a function's domain to evaluate
# next, drawing on previously observed values to make optimal decisions. In order to do
# this effectively, we need a way of representing our uncertainty about the black-box
# function $f$, which we can update in light of observing more data. Gaussian processes
# will be an ideal tool for this purpose!
#
# *Surrogate models* lie at the heart of BO, and are used to model the black-box
# function. GPs are a natural choice for this model, as they not only provide point
# estimates for the values taken by the function throughout its domain, but crucially
# provide a full predictive posterior *distribution* of the range of values the function
# may take. This rich quantification of uncertainty enables BO to balance *exploration*
# and *exploitation* in order to efficiently converge upon minima.
#
# Having chosen a surrogate model, which we can use to express our current beliefs about
# the black-box function, ideally we would like a method which can use the surrogate
# model's posterior distribution to automatically decide which point(s) in the black-box
# function's domain to query next. This is where *acquisition functions* come in. The
# acquisition function $\alpha: X \to \mathbb{R}$ is defined over the same domain as the
# surrogate model, and uses the surrogate model's posterior distribution to quantify the
# expected *utility*, $U$, of evaluating the black-box function at a given point. Simply
# put, for each point in the black-box function's domain, $\mathbf{x} \in X$, the
# acquisition function quantifies how useful it would be to evaluate the black-box
# function at $\mathbf{x}$ in order to find the minimum of the black-box function, whilst
# taking into consideration all the datapoints observed so far. Therefore, in order to
# decide which point to query next we simply choose the point which maximises the
# acquisition function, using an optimiser such as L-BFGS ([Liu and Nocedal,
# 1989](https://link.springer.com/article/10.1007/BF01589116)).
#
# The Bayesian optimisation loop can be summarised as follows, with $i$ denoting the
# current iteration:
#
# 1. Select the next point to query, $\mathbf{x}_{i}$, by maximising the acquisition function $\alpha$, defined using the surrogate model $\mathcal{M}_i$ conditioned on previously observed data $\mathcal{D}_i$:
#
# $$\mathbf{x}_{i} = \arg\max_{\mathbf{x}} \alpha (\mathbf{x}; \mathcal{D}_i,
# \mathcal{M}_i)$$
#
# 2. Evaluate the objective function at $\mathbf{x}_i$, yielding observation $y_i =
#    f(\mathbf{x}_i)$.
#
# 3. Append the most recent observation to the dataset, $\mathcal{D}_{i+1} = \mathcal{D}_i
#    \cup \{(\mathbf{x}_i, y_i)\}$.
#
# 4. Condition the model on the updated dataset to yield $\mathcal{M}_{i+1}$.
#
# This process is repeated until some stopping criterion is met, such as a function
# evaluation budget being exhausted.
#
# There are a plethora of acquisition functions to choose from, each with their own
# advantages and disadvantages, of which ([Shahriari et al., 2015](https://www.cs.ox.ac.uk/people/nando.defreitas/publications/BayesOptLoop.pdf))
# provides an excellent overview.
#
# In this guide we will focus on *Thompson sampling*, a conceptually simple yet effective
# method for characterising the utility of querying points in a black-box function's
# domain, which will be useful in demonstrating the key aspects of BO.

# %% [markdown]
# ## Thompson Sampling
#
# Thompson sampling ([Thompson, 1933](https://www.dropbox.com/s/yhn9prnr5bz0156/1933-thompson.pdf)) is a simple method which
# naturally balances exploration and exploitation. The core idea is to, at each iteration
# of the BO loop, sample a function, $g$, from the posterior distribution of the surrogate
# model $\mathcal{M}_i$, and then evaluate the black-box function at the point(s) which
# minimise this sample. Given a sample $g$, from the posterior distribution given by the model $\mathcal{M}_i$ the Thompson sampling utility function is defined as:
#
# $$U_{\text{TS}}(\mathbf{x}; \mathcal{D}_i, \mathcal{M}_i) = - g(\mathbf{x})$$
#
# Note the negative sign; this is included as we want to maximise the *utility* of
# evaluating the black-box function $f$ at a given point. We interested in finding the
# minimum of $f$, so we maximise the negative of the sample from the posterior distribution $g$.
#
# As a toy example, we shall be applying BO to the widely used [Forrester
# function](https://www.sfu.ca/~ssurjano/forretal08.html):
#
# $$f(x) = (6x - 2)^2 \sin(12x - 4)$$
#
# treating $f$ as a black-box function. Moreover, we shall restrict the domain of the
# function to $\mathbf{x} \in [0, 1]$. We shall also *standardise* the output of the function, such that
# it has a mean of 0 and standard deviation of 1. This is quite common practice when using
# GPs; we're using a zero mean prior, so ensuring that our data has a mean of zero aligns
# with this, and often we have scale parameters in the covariance function, which
# are frequently initialised, or have priors set on them, under the assumption that the
# function being modelled has unit variance. For similar reasons, it can also be useful
# to normalise the inputs to a GP. The global minimum of this (standardised)
# function is located at $x = 0.757$, where $f(x) = -1.463$.


# %%
def standardised_forrester(x: Float[Array, "N 1"]) -> Float[Array, "N 1"]:
    mean = 0.45321
    std = 4.4258
    return ((6 * x - 2) ** 2 * jnp.sin(12 * x - 4) - mean) / std


# %% [markdown]
# We'll first go through one iteration of the BO loop step-by-step, before wrapping this
# up in a loop to perform the full optimisation.

# %% [markdown]
# First we'll specify the domain over which we wish to optimise the function, as well as
# sampling some initial points for fitting our surrogate model using a space-filling design.

# %%
lower_bound = jnp.array([0.0])
upper_bound = jnp.array([1.0])
initial_sample_num = 5

initial_x = tfp.mcmc.sample_halton_sequence(
    dim=1, num_results=initial_sample_num, seed=key, dtype=jnp.float64
).reshape(-1, 1)
initial_y = standardised_forrester(initial_x)
D = gpx.Dataset(X=initial_x, y=initial_y)


# %% [markdown]
# Next we'll define our GP model in the usual way, using a Matérn52 kernel, and fit the
# kernel parameters by minimising the negative log-marginal likelihood. We'll wrap this in
# a function as we'll be repeating this process at each iteration of the BO loop.


# %%


def return_optimised_posterior(
    data: gpx.Dataset, prior: gpx.gps.Prior, key: Array
) -> gpx.gps.AbstractPosterior:
    # Our function is noise-free, so we set the observation noise's standard deviation to a very small value
    likelihood = gpx.likelihoods.Gaussian(
        num_datapoints=data.n, obs_stddev=Static(jnp.array(1e-6))
    )

    posterior = prior * likelihood

    opt_posterior, _ = gpx.fit(
        model=posterior,
        objective=lambda p, d: -gpx.objectives.conjugate_mll(p, d),
        train_data=data,
        optim=ox.adam(learning_rate=0.01),
        num_iters=1000,
        safe=True,
        key=key,
        verbose=False,
    )

    return opt_posterior


mean = gpx.mean_functions.Zero()
kernel = gpx.kernels.Matern52(n_dims=1)
prior = gpx.gps.Prior(mean_function=mean, kernel=kernel)
opt_posterior = return_optimised_posterior(D, prior, key)

# %% [markdown]
# We can then sample a function from the posterior distribution of the surrogate model. We
# will do this using the `sample_approx` method, which generates an approximate sample
# from the posterior using decoupled sampling introduced in ([Wilson et al.,
# 2020](https://proceedings.mlr.press/v119/wilson20a.html)) and discussed in our [Pathwise
# Sampling Notebook](https://docs.jaxgaussianprocesses.com/examples/spatial/). This method
# is used as it enables us to sample from the posterior in a manner which scales linearly
# with the number of points sampled, $O(N)$, mitigating the cubic cost associated with
# drawing exact samples from a GP posterior, $O(N^3)$. It also generates more accurate
# samples than many other methods for drawing approximate samples from a GP posterior.
#
# Note that we also define a `utility_fn` which calls the approximate
# sample but returns the value returned as a scalar. This is because the `sample_approx`
# function returns an array of shape $[N, B]$, with $N$ being the number of points within
# each sample and $B$ being the number of samples drawn. We'll only be drawing (and
# optimising) one sample at a time, and our optimiser requires the function being
# optimised to return a scalar output (only querying it at $N=1$ points), so we'll remove the axes from the returned value.

# %%
approx_sample = opt_posterior.sample_approx(
    num_samples=1, train_data=D, key=key, num_features=500
)
utility_fn = lambda x: approx_sample(x)[0][0]


# %% [markdown]
# In order to minimise the sample, we'll be using the L-BFGS-B ([Byrd et al., 1995](https://epubs.siam.org/doi/abs/10.1137/0916069)) optimiser from the `jaxopt`
# library. This is a gradient-based optimiser which performs optimisation within a bounded
# domain. In order to perform optimisation, this optimiser requires a point to start from.
# Therefore, we will first query our sample from the posterior at a random set of points,
# and then use the lowest point from this set of points as the starting point for the
# optimiser. In this example we'll sample 100 points from the posterior, due to the simple
# nature of the Forrester function. However, in practice it can be beneficial to
# adopt a more sophisticated approach, and there are several heuristics available in the
# literature (see for example ([Le Riche and Picheny,
# 2021](https://arxiv.org/abs/2103.16649))). For instance, one may randomly sample the
# posterior at a number of points proportional to the dimensionality of the input space,
# and one may run gradient-based optimisation from multiple of these points, to reduce the
# risk of converging upon local minima.


# %%
def optimise_sample(
    sample: FunctionalSample,
    key: Int[Array, ""],
    lower_bound: Float[Array, "D"],
    upper_bound: Float[Array, "D"],
    num_initial_sample_points: int,
) -> ScalarFloat:
    initial_sample_points = jr.uniform(
        key,
        shape=(num_initial_sample_points, lower_bound.shape[0]),
        dtype=jnp.float64,
        minval=lower_bound,
        maxval=upper_bound,
    )
    initial_sample_y = sample(initial_sample_points)
    best_x = jnp.array([initial_sample_points[jnp.argmin(initial_sample_y)]])

    # We want to maximise the utility function, but the optimiser performs minimisation. Since we're minimising the sample drawn, the sample is actually the negative utility function.
    negative_utility_fn = lambda x: sample(x)[0][0]
    lbfgsb = ScipyBoundedMinimize(fun=negative_utility_fn, method="l-bfgs-b")
    bounds = (lower_bound, upper_bound)
    x_star = lbfgsb.run(best_x, bounds=bounds).params
    return x_star


x_star = optimise_sample(approx_sample, key, lower_bound, upper_bound, 100)
y_star = standardised_forrester(x_star)


# %% [markdown]
# Having found the minimum of the sample from the posterior, we can then evaluate the
# black-box objective function at this point, and append the new observation to our dataset.
#
# Below we plot the posterior distribution of the surrogate model, along with the sample
# drawn from the model, and the minimiser of this sample returned from the optimiser,
# which we denote with a star.


# %%
def plot_bayes_opt(
    posterior: gpx.gps.AbstractPosterior,
    sample: FunctionalSample,
    dataset: gpx.Dataset,
    queried_x: ScalarFloat,
) -> None:
    plt_x = jnp.linspace(0, 1, 1000).reshape(-1, 1)
    forrester_y = standardised_forrester(plt_x)
    sample_y = sample(plt_x)

    latent_dist = posterior.predict(plt_x, train_data=dataset)
    predictive_dist = posterior.likelihood(latent_dist)

    predictive_mean = predictive_dist.mean()
    predictive_std = predictive_dist.stddev()

    fig, ax = plt.subplots()
    ax.plot(plt_x, predictive_mean, label="Predictive Mean", color=cols[1])
    ax.fill_between(
        plt_x.squeeze(),
        predictive_mean - 2 * predictive_std,
        predictive_mean + 2 * predictive_std,
        alpha=0.2,
        label="Two sigma",
        color=cols[1],
    )
    ax.plot(
        plt_x,
        predictive_mean - 2 * predictive_std,
        linestyle="--",
        linewidth=1,
        color=cols[1],
    )
    ax.plot(
        plt_x,
        predictive_mean + 2 * predictive_std,
        linestyle="--",
        linewidth=1,
        color=cols[1],
    )
    ax.plot(plt_x, sample_y, label="Posterior Sample")
    ax.plot(
        plt_x,
        forrester_y,
        label="Forrester Function",
        color=cols[0],
        linestyle="--",
        linewidth=2,
    )
    ax.axvline(x=0.757, linestyle=":", color=cols[3], label="True Optimum")
    ax.scatter(dataset.X, dataset.y, label="Observations", color=cols[2], zorder=2)
    ax.scatter(
        queried_x,
        sample(queried_x),
        label="Posterior Sample Optimum",
        marker="*",
        color=cols[3],
        zorder=3,
    )
    ax.legend(loc="center left", bbox_to_anchor=(0.975, 0.5))
    plt.show()


plot_bayes_opt(opt_posterior, approx_sample, D, x_star)

# %% [markdown]
# At this point we can update our model with the newly augmented dataset, and repeat the
# whole process until some stopping criterion is met. Below we repeat this process for 10
# iterations, printing out the queried point and the value of the black-box function at
# each iteration.

# %%
bo_iters = 5

# Set up initial dataset
initial_x = tfp.mcmc.sample_halton_sequence(
    dim=1, num_results=initial_sample_num, seed=key, dtype=jnp.float64
).reshape(-1, 1)
initial_y = standardised_forrester(initial_x)
D = gpx.Dataset(X=initial_x, y=initial_y)

for _ in range(bo_iters):
    key, subkey = jr.split(key)

    # Generate optimised posterior using previously observed data
    mean = gpx.mean_functions.Zero()
    kernel = gpx.kernels.Matern52(n_dims=1)
    prior = gpx.gps.Prior(mean_function=mean, kernel=kernel)
    opt_posterior = return_optimised_posterior(D, prior, subkey)

    # Draw a sample from the posterior, and find the minimiser of it
    approx_sample = opt_posterior.sample_approx(
        num_samples=1, train_data=D, key=subkey, num_features=500
    )
    x_star = optimise_sample(
        approx_sample, subkey, lower_bound, upper_bound, num_initial_sample_points=100
    )

    plot_bayes_opt(opt_posterior, approx_sample, D, x_star)

    # Evaluate the black-box function at the best point observed so far, and add it to the dataset
    y_star = standardised_forrester(x_star)
    print(f"Queried Point: {x_star}, Black-Box Function Value: {y_star}")
    D = D + gpx.Dataset(X=x_star, y=y_star)

# %% [markdown]
# Below we plot the best observed black-box function value against the number of times
# the black-box function has been evaluated. Note that the first 5 samples are randomly
# sampled to fit the initial GP model, and we denote the start of using BO to sample with
# the dotted vertical line.
#
# We can see that the BO algorithm quickly converges to the global minimum of the
# black-box function!
#

# %%
fig, ax = plt.subplots()
fn_evaluations = jnp.arange(1, bo_iters + initial_sample_num + 1)
cumulative_best_y = jax.lax.associative_scan(jax.numpy.minimum, D.y)
ax.plot(fn_evaluations, cumulative_best_y)
ax.axvline(x=initial_sample_num, linestyle=":")
ax.axhline(y=-1.463, linestyle="--", label="True Minimum")
ax.set_xlabel("Number of Black-Box Function Evaluations")
ax.set_ylabel("Best Observed Value")
ax.legend()
plt.show()


# %% [markdown]
# ### A More Challenging Example - The Six-Hump Camel Function

# %% [markdown]
# We'll now apply BO to a more challenging example, the [Six-Hump Camel
# Function](https://www.sfu.ca/~ssurjano/camel6.html). This is a function of two inputs
# defined as follows:
#
# $$f(x_1, x_2) = (4 - 2.1x_1^2 + \frac{x_1^4}{3})x_1^2 + x_1x_2 + (-4 + 4x_2^2)x_2^2$$
#
# We'll be evaluating it over the domain $x_1 \in [-2, 2]$ and $x_2 \in [-1, 1]$, and
# shall standardise it. The
# global minima of this function are located at $\mathbf{x} = (0.0898, -0.7126)$ and
# $\mathbf{x} = (-0.0898, 0.7126)$, where the standardised function takes the value $f(\mathbf{x}) =
# -1.8377$.


# %%
def standardised_six_hump_camel(x: Float[Array, "N 2"]) -> Float[Array, "N 1"]:
    mean = 1.12767
    std = 1.17500
    x1 = x[..., :1]
    x2 = x[..., 1:]
    term1 = (4 - 2.1 * x1**2 + x1**4 / 3) * x1**2
    term2 = x1 * x2
    term3 = (-4 + 4 * x2**2) * x2**2
    return (term1 + term2 + term3 - mean) / std


# %% [markdown]
# First, we'll visualise the function over the domain of interest:

# %%
x1 = jnp.linspace(-2, 2, 100)
x2 = jnp.linspace(-1, 1, 100)
x1, x2 = jnp.meshgrid(x1, x2)
x = jnp.stack([x1.flatten(), x2.flatten()], axis=1)
y = standardised_six_hump_camel(x)

fig, ax = plt.subplots(subplot_kw={"projection": "3d"})
surf = ax.plot_surface(
    x1,
    x2,
    y.reshape(x1.shape[0], x2.shape[0]),
    linewidth=0,
    cmap=cm.coolwarm,
    antialiased=False,
)
ax.set_xlabel("x1")
ax.set_ylabel("x2")
plt.show()

# %% [markdown]
# For more clarity, we can generate a contour plot of the function which enables us to see
# the global minima of the function more clearly.

# %%
x_star_one = jnp.array([[0.0898, -0.7126]])
x_star_two = jnp.array([[-0.0898, 0.7126]])
fig, ax = plt.subplots()
contour_plot = ax.contourf(
    x1, x2, y.reshape(x1.shape[0], x2.shape[0]), cmap=cm.coolwarm, levels=40
)
ax.scatter(
    x_star_one[0][0], x_star_one[0][1], marker="*", color=cols[2], label="Global Minima"
)
ax.scatter(x_star_two[0][0], x_star_two[0][1], marker="*", color=cols[2])
ax.set_xlabel("x1")
ax.set_ylabel("x2")
fig.colorbar(contour_plot)
ax.legend()
plt.show()

# %% [markdown]
# Next, we'll run the BO loop using Thompson sampling as before. This time we'll run the
# experiment 5 times in order to see how the algorithm performs on average, with different
# starting points for the initial GP model. This is good practice, as the performance
# obtained is likely to vary between runs depending on the initialisation samples used to
# fit the initial GP model.

# %%
lower_bound = jnp.array([-2.0, -1.0])
upper_bound = jnp.array([2.0, 1.0])
initial_sample_num = 5
bo_iters = 12
num_experiments = 5
bo_experiment_results = []

for experiment in range(num_experiments):
    print(f"Starting Experiment: {experiment + 1}")
    # Set up initial dataset
    initial_x = tfp.mcmc.sample_halton_sequence(
        dim=2, num_results=initial_sample_num, seed=key, dtype=jnp.float64
    )
    initial_x = jnp.array(lower_bound + (upper_bound - lower_bound) * initial_x)
    initial_y = standardised_six_hump_camel(initial_x)
    D = gpx.Dataset(X=initial_x, y=initial_y)

    for i in range(bo_iters):
        key, subkey = jr.split(key)

        # Generate optimised posterior
        mean = gpx.mean_functions.Zero()
        kernel = gpx.kernels.Matern52(
            active_dims=[0, 1], lengthscale=jnp.array([1.0, 1.0]), variance=2.0
        )
        prior = gpx.gps.Prior(mean_function=mean, kernel=kernel)
        opt_posterior = return_optimised_posterior(D, prior, subkey)

        # Draw a sample from the posterior, and find the minimiser of it
        approx_sample = opt_posterior.sample_approx(
            num_samples=1, train_data=D, key=subkey, num_features=500
        )
        x_star = optimise_sample(
            approx_sample,
            subkey,
            lower_bound,
            upper_bound,
            num_initial_sample_points=1000,
        )

        # Evaluate the black-box function at the best point observed so far, and add it to the dataset
        y_star = standardised_six_hump_camel(x_star)
        print(
            f"BO Iteration: {i + 1}, Queried Point: {x_star}, Black-Box Function Value:"
            f" {y_star}"
        )
        D = D + gpx.Dataset(X=x_star, y=y_star)
    bo_experiment_results.append(D)


# %% [markdown]
# We'll also run a random benchmark, whereby we randomly sample from the search space for
# 20 iterations. This is a useful benchmark to compare the performance of BO against in
# order to ascertain how much of an advantage BO provides over such a simple approach.
#

# %%
random_experiment_results = []
for i in range(num_experiments):
    key, subkey = jr.split(key)
    initial_x = bo_experiment_results[i].X[:5]
    initial_y = bo_experiment_results[i].y[:5]
    final_x = jr.uniform(
        key,
        shape=(bo_iters, 2),
        dtype=jnp.float64,
        minval=lower_bound,
        maxval=upper_bound,
    )
    final_y = standardised_six_hump_camel(final_x)
    random_x = jnp.concatenate([initial_x, final_x], axis=0)
    random_y = jnp.concatenate([initial_y, final_y], axis=0)
    random_experiment_results.append(gpx.Dataset(X=random_x, y=random_y))


# %% [markdown]
# Finally, we'll process the experiment results to find the log regret at each iteration
# of the experiments. The regret is defined as the difference between the minimum value of
# the black-box function observed so far and the true global minimum of the black box
# function. Mathematically, at time $t$, with observations $\mathcal{D}_t$, for function
# $f$ with global minimum $f^*$, the regret is defined as:
#
# $$\text{regret}_t = \min_{\mathbf{x} \in \mathcal{D_t}}f(\mathbf{x}) - f^*$$
#
# We'll then take the mean and standard deviation of the log of the regret values across
# the 5 experiments.


# %%
def obtain_log_regret_statistics(
    experiment_results: List[gpx.Dataset],
    global_minimum: ScalarFloat,
) -> Tuple[Float[Array, "N 1"], Float[Array, "N 1"]]:
    log_regret_results = []
    for exp_result in experiment_results:
        observations = exp_result.y
        cumulative_best_observations = jax.lax.associative_scan(
            jax.numpy.minimum, observations
        )
        regret = cumulative_best_observations - global_minimum
        log_regret = jnp.log(regret)
        log_regret_results.append(log_regret)

    log_regret_results = jnp.array(log_regret_results)
    log_regret_mean = jnp.mean(log_regret_results, axis=0)
    log_regret_std = jnp.std(log_regret_results, axis=0)
    return log_regret_mean, log_regret_std


bo_log_regret_mean, bo_log_regret_std = obtain_log_regret_statistics(
    bo_experiment_results, -1.8377
)
(
    random_log_regret_mean,
    random_log_regret_std,
) = obtain_log_regret_statistics(random_experiment_results, -1.8377)

# %% [markdown]
# Now, when we plot the mean and standard deviation of the log regret at each iteration,
# we can see that BO outperforms random sampling!

# %%
fig, ax = plt.subplots()
fn_evaluations = jnp.arange(1, bo_iters + initial_sample_num + 1)
ax.plot(fn_evaluations, bo_log_regret_mean, label="Bayesian Optimisation")
ax.fill_between(
    fn_evaluations,
    bo_log_regret_mean[:, 0] - bo_log_regret_std[:, 0],
    bo_log_regret_mean[:, 0] + bo_log_regret_std[:, 0],
    alpha=0.2,
)
ax.plot(fn_evaluations, random_log_regret_mean, label="Random Search")
ax.fill_between(
    fn_evaluations,
    random_log_regret_mean[:, 0] - random_log_regret_std[:, 0],
    random_log_regret_mean[:, 0] + random_log_regret_std[:, 0],
    alpha=0.2,
)
ax.axvline(x=initial_sample_num, linestyle=":")
ax.set_xlabel("Number of Black-Box Function Evaluations")
ax.set_ylabel("Log Regret")
ax.legend()
plt.show()

# %% [markdown]
# It can also be useful to plot the queried points over the course of a single BO run, in
# order to gain some insight into how the algorithm queries the search space. Below
# we do this for one of the BO experiments, and can see that the algorithm initially
# performs some exploration of the search space whilst it is uncertain about the black-box
# function, but it then hones in one one of the global minima of the function, as we would hope!

# %%
fig, ax = plt.subplots()
contour_plot = ax.contourf(
    x1, x2, y.reshape(x1.shape[0], x2.shape[0]), cmap=cm.coolwarm, levels=40
)
ax.scatter(
    x_star_one[0][0],
    x_star_one[0][1],
    marker="*",
    color=cols[2],
    label="Global Minimum",
    zorder=2,
)
ax.scatter(x_star_two[0][0], x_star_two[0][1], marker="*", color=cols[2], zorder=2)
ax.scatter(
    bo_experiment_results[1].X[:, 0],
    bo_experiment_results[1].X[:, 1],
    marker="x",
    color=cols[1],
    label="Bayesian Optimisation Queries",
)
ax.set_xlabel("x1")
ax.set_ylabel("x2")
fig.colorbar(contour_plot)
ax.legend()
plt.show()

# %% [markdown]
# ### Other Acquisition Functions and Further Reading
#
# As mentioned previously, there are many acquisition functions which one may use to
# characterise the expected utility of querying the black-box function at a given point.
# We list two of the most popular below:
#
# - **Probability of Improvement (PI)** ([Kushner, 1964](https://asmedigitalcollection.asme.org/fluidsengineering/article/86/1/97/392213/A-New-Method-of-Locating-the-Maximum-Point-of-an)): Given the lowest objective function observation
#   so far, $f(\mathbf{x}^*)$, PI calculates the probability that the objective function's
#   value at a given point $\mathbf{x}$ is lower than $f(\mathbf{x}^*)$. Given a GP
#   surrogate model $\mathcal{M}_i$, PI is defined mathematically as:
#   $$
#   \alpha_{\text{PI}}(\mathbf{x}; \mathcal{D}_i, \mathcal{M}_i) = \mathbb{P}[\mathcal{M}_i (\mathbf{x}) < f(\mathbf{x}^*)] = \Phi \left(\frac{f(\mathbf{x}^*) - \mu_{\mathcal{M}_i}(\mathbf{x})}{\sigma_{\mathcal{M}_i}(\mathbf{x})}\right)
#   $$
#
#   with $\Phi(\cdot)$ denoting the standard normal cumulative distribution function.
#
# - **Expected Improvement (EI)** ([Močkus, 1974](https://link.springer.com/chapter/10.1007/3-540-07165-2_55)) - EI goes beyond PI by not only considering the
#   probability of improving on the current best observed point, but also taking into
#   account the *magnitude* of improvement. Mathematically, this is defined as
#   follows:
#   $$
#   \begin{aligned}
#   \alpha_{\text{EI}}(\mathbf{x};\mathcal{D}_i, \mathcal{M}_i) &= \mathbb{E}[(f(\mathbf{x}^*) - \mathcal{M}_i(\mathbf{x}))\mathbb{I}(\mathcal{M}_i(\mathbf{x}) < f(\mathbf{x}^*))] \\
#   &= \underbrace{(f(\mathbf{x}^*) - \mu_{\mathcal{M}_i}(\mathbf{x}))\Phi
#   \left(\frac{f(\mathbf{x}^*) -
#   \mu_{\mathcal{M}_i}(\mathbf{x})}{\sigma_{\mathcal{M}_i}(\mathbf{x})}\right)}_\text{exploits
#   areas with low mean} \\
#   &+  \underbrace{\sigma_{\mathcal{M}_i}(\mathbf{x}) \phi \left(\frac{f(\mathbf{x}^*) - \mu_{\mathcal{M}_i}(\mathbf{x})}{\sigma_{\mathcal{M}_i}(\mathbf{x})}\right)}_\text{explores areas with high variance} \nonumber
#   \end{aligned}
#   $$
#
#   with $\mathbb{I}(\cdot)$ denoting the indicator function and $\phi(\cdot)$ being the
#   standard normal probability density function.
#
# For those particularly interested in diving deeper into Bayesian optimisation, be sure
# to check out Shahriari et al.'s "[Taking the Human Out of the Loop:
# A Review of Bayesian
# Optimization](https://www.cs.ox.ac.uk/people/nando.defreitas/publications/BayesOptLoop.pdf)",
# which includes a wide variety of acquisition functions, as well as some examples of more
# exotic BO problems, such as problems which also feature unknown constraints.
#
# ## System Configuration

# %%
# %reload_ext watermark
# %watermark -n -u -v -iv -w -a 'Thomas Christie'



# Copyright 2023 The GPJax Contributors. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ==============================================================================

from beartype.typing import (
    Mapping,
    Optional,
)
import jax.numpy as jnp

from gpjax.dataset import Dataset
from gpjax.decision_making.test_functions import Quadratic
from gpjax.decision_making.utility_functions import (
    AbstractSinglePointUtilityFunctionBuilder,
    SinglePointUtilityFunction,
)
from gpjax.gps import (
    ConjugatePosterior,
    NonConjugatePosterior,
    Prior,
)
from gpjax.kernels import RBF
from gpjax.likelihoods import (
    Gaussian,
    Poisson,
)
from gpjax.mean_functions import (
    AbstractMeanFunction,
    Zero,
)
from gpjax.typing import KeyArray


class QuadraticSinglePointUtilityFunctionBuilder(
    AbstractSinglePointUtilityFunctionBuilder
):
    """
    Dummy utility function builder for testing purposes, which returns the negative
    of the value of a quadratic test function at the input points. This is because
    utility functions are *maximised*, and we wish to *minimise* the quadratic test
    function. Note that this is a `SinglePointUtilityFunctionBuilder`.
    """

    def build_utility_function(
        self,
        posteriors: Mapping[str, ConjugatePosterior],
        datasets: Mapping[str, Dataset],
        key: KeyArray,
    ) -> SinglePointUtilityFunction:
        test_function = Quadratic()
        return lambda x: -1.0 * test_function.evaluate(
            x
        )  # Utility functions are *maximised*


def generate_dummy_conjugate_posterior(
    dataset: Dataset,
    mean_function: Optional[AbstractMeanFunction] = None,
) -> ConjugatePosterior:
    kernel = RBF(lengthscale=jnp.ones(dataset.X.shape[1]))
    if mean_function is None:
        mean_function = Zero()
    prior = Prior(kernel=kernel, mean_function=mean_function)
    likelihood = Gaussian(num_datapoints=dataset.n, obs_stddev=1e-6)
    posterior = prior * likelihood
    return posterior


def generate_dummy_non_conjugate_posterior(
    dataset: Dataset,
    mean_function: Optional[AbstractMeanFunction] = None,
) -> NonConjugatePosterior:
    kernel = RBF(lengthscale=jnp.ones(dataset.X.shape[1]))
    if mean_function is None:
        mean_function = Zero()
    prior = Prior(kernel=kernel, mean_function=mean_function)
    likelihood = Poisson(num_datapoints=dataset.n)
    posterior = prior * likelihood
    return posterior



# ---
# jupyter:
#   jupytext:
#     cell_metadata_filter: -all
#     custom_cell_magics: kql
#     text_representation:
#       extension: .py
#       format_name: percent
#       format_version: '1.3'
#       jupytext_version: 1.16.4
#   kernelspec:
#     display_name: gpjax
#     language: python
#     name: python3
# ---

# %% [markdown]
# # Gaussian Processes for Vector Fields and Ocean Current Modelling
#
# In this notebook, we use Gaussian processes to learn vector-valued functions. We will be
# recreating the results by [Berlinghieri et al. (2023)](https://arxiv.org/pdf/2302.10364.pdf) by an
# application to real-world ocean surface velocity data, collected via surface drifters.
#
# Surface drifters are measurement devices that measure the dynamics and circulation patterns of the world's oceans. Studying and predicting ocean currents are important to climate research, for example, forecasting and predicting oil spills, oceanographic surveying of eddies and upwelling, or providing information on the distribution of biomass in ecosystems. We will be using the [Gulf Drifters Open dataset](https://zenodo.org/record/4421585), which contains all publicly available surface drifter trajectories from the Gulf of Mexico spanning 28 years.
# %%
from dataclasses import (
    dataclass,
    field,
)

from jax import (
    config,
    hessian,
)
import jax.numpy as jnp
import jax.random as jr
from jaxtyping import (
    Array,
    Float,
    install_import_hook,
)
from matplotlib import rcParams
import matplotlib.pyplot as plt
import pandas as pd
import tensorflow_probability as tfp

from examples.utils import use_mpl_style
from gpjax.kernels.computations import DenseKernelComputation

config.update("jax_enable_x64", True)


with install_import_hook("gpjax", "beartype.beartype"):
    import gpjax as gpx


# set the default style for plotting
use_mpl_style()

key = jr.key(42)

colors = rcParams["axes.prop_cycle"].by_key()["color"]


# %% [markdown]
# ## Data loading and preprocessing
# The real dataset has been binned into an $N=34\times16$ grid, equally spaced over the longitude-latitude interval $[-90.8,-83.8] \times [24.0,27.5]$. Each bin has a size $\approx 0.21\times0.21$, and contains the average velocity across all measurements that fall inside it.
#
# We will call this binned ocean data the ground truth, and label it with the vector field
# $$
# \mathbf{F} \equiv \mathbf{F}(\mathbf{x}),
# $$
# where $\mathbf{x} = (x^{(0)}$,$x^{(1)})^\text{T}$, with a vector basis in the standard Cartesian directions (dimensions will be indicated by superscripts).
#
# We shall label the ground truth $D_0=\left\{ \left(\mathbf{x}_{0,i} , \mathbf{y}_{0,i} \right)\right\}_{i=1}^N$, where $\mathbf{y}_{0,i}$ is the 2-dimensional velocity vector at the $i$-th location, $\mathbf{x}_{0,i}$. The training dataset contains simulated measurements from ocean drifters $D_T=\left\{\left(\mathbf{x}_{T,i}, \mathbf{y}_{T,i} \right)\right\}_{i=1}^{N_T}$, $N_T = 20$ in this case (the subscripts indicate the ground truth and the simulated measurements respectively).
#


# %%
# function to place data from csv into correct array shape
def prepare_data(df):
    pos = jnp.array([df["lon"], df["lat"]])
    vel = jnp.array([df["ubar"], df["vbar"]])
    # extract shape stored as 'metadata' in the test data
    try:
        shape = (int(df["shape"][1]), int(df["shape"][0]))  # shape = (34,16)
        return pos, vel, shape
    except KeyError:
        return pos, vel


# loading in data

gulf_data_train = pd.read_csv(
    "https://raw.githubusercontent.com/JaxGaussianProcesses/static/main/data/gulfdata_train.csv"
)
gulf_data_test = pd.read_csv(
    "https://raw.githubusercontent.com/JaxGaussianProcesses/static/main/data/gulfdata_test.csv"
)


pos_test, vel_test, shape = prepare_data(gulf_data_test)
pos_train, vel_train = prepare_data(gulf_data_train)

fig, ax = plt.subplots(1, 1, figsize=(6, 3))
ax.quiver(
    pos_test[0],
    pos_test[1],
    vel_test[0],
    vel_test[1],
    color=colors[0],
    label="Ocean Current",
    angles="xy",
    scale=10,
)
ax.quiver(
    pos_train[0],
    pos_train[1],
    vel_train[0],
    vel_train[1],
    color=colors[1],
    alpha=0.7,
    label="Drifter",
    angles="xy",
    scale=10,
)

ax.set(
    xlabel="Longitude",
    ylabel="Latitude",
)
ax.legend(
    framealpha=0.0,
    ncols=2,
    fontsize="medium",
    bbox_to_anchor=(0.5, -0.3),
    loc="lower center",
)
plt.show()

# %% [markdown]
# ## Problem Setting
# We aim to obtain estimates for $\mathbf{F}$ at the set of points $\left\{ \mathbf{x}_{0,i} \right\}_{i=1}^N$ using Gaussian processes, followed by a comparison of the latent model to the ground truth $D_0$. Note that $D_0$ is not passed into any functions used  by GPJax, and is only used to compare against the two GP models at the end of the notebook.
#
# Since $\mathbf{F}$ is a vector-valued function, we require GPs that can directly learn vector-valued functions[<sup>1</sup>](#fn1). To implement this in GPJax, the problem can be changed to learn a scalar-valued function by 'massaging' the data into a  $2N\times2N$ problem, such that each dimension of our GP is associated with a *component* of $\mathbf{y}_{T,i}$.
#
# For a particular measurement $\mathbf{y}$ (training or testing) at location $\mathbf{x}$, the components $(y^{(0)}, y^{(1)})$ are described by the latent vector field $\mathbf{F}$, such that
#
# $$
# \mathbf{y} = \mathbf{F}(\mathbf{x}) = \left(\begin{array}{l}
# f^{(0)}\left(\mathbf{x}\right) \\
# f^{(1)}\left(\mathbf{x}\right)
# \end{array}\right),
# $$
#
# where each $f^{(z)}\left(\mathbf{x}\right), z \in \{0,1\}$ is a scalar-valued function.
#
# Now consider the scalar-valued function $g: \mathbb{R}^2 \times\{0,1\} \rightarrow \mathbb{R}$, such that
#
# $$
# g \left(\mathbf{x} , 0 \right) = f^{(0)} ( \mathbf{x} ), \text{and } g \left( \mathbf{x}, 1 \right)=f^{(1)}\left(\mathbf{x}\right).
# $$
#
# We have increased the input dimension by 1, from the 2D $\mathbf{x}$ to the 3D $\mathbf{X} = \left(\mathbf{x}, 0\right)$ or $\mathbf{X} = \left(\mathbf{x}, 1\right)$.
#
# By choosing the value of the third dimension, 0 or 1, we may now incorporate this
# information into the computation of the kernel.
# We therefore make new 3D datasets $D_{T,3D} = \left\{\left( \mathbf{X}_{T,i},\mathbf{Y}_{T,i} \right) \right\} _{i=0}^{2N_T}$ and $D_{0,3D} = \left\{\left( \mathbf{X}_{0,i},\mathbf{Y}_{0,i} \right) \right\} _{i=0}^{2N}$ that incorporates this new labelling, such that for each dataset (indicated by the subscript $D = 0$ or $D=T$),
#
# $$
# X_{D,i} = \left( \mathbf{x}_{D,i}, z \right),
# $$
# and
# $$
# Y_{D,i} = y_{D,i}^{(z)},
# $$
#
# where $z = 0$ if $i$ is odd and $z=1$ if $i$ is even.


# %%
# Change vectors x -> X = (x,z), and vectors y -> Y = (y,z) via the artificial z label
def label_position(data):
    # introduce alternating z label
    n_points = len(data[0])
    label = jnp.tile(jnp.array([0.0, 1.0]), n_points)
    return jnp.vstack((jnp.repeat(data, repeats=2, axis=1), label)).T


# change vectors y -> Y by reshaping the velocity measurements
def stack_velocity(data):
    return data.T.flatten().reshape(-1, 1)


def dataset_3d(pos, vel):
    return gpx.Dataset(label_position(pos), stack_velocity(vel))


# label and place the training data into a Dataset object to be used by GPJax
dataset_train = dataset_3d(pos_train, vel_train)

# we also require the testing data to be relabelled for later use, such that we can query the 2Nx2N GP at the test points
dataset_ground_truth = dataset_3d(pos_test, vel_test)


# %% [markdown]
# ## Velocity (dimension) decomposition
# Having labelled the data, we are now in a position to use GPJax to learn the function $g$, and hence $\mathbf{F}$. A naive approach to the problem is to apply a GP prior directly to the velocities of each dimension independently, which is called the *velocity* GP. For our prior, we choose an isotropic mean 0 over all dimensions of the GP, and a piecewise kernel that depends on the $z$ labels of the inputs, such that for two inputs $\mathbf{X} = \left( \mathbf{x}, z \right )$ and $\mathbf{X}^\prime = \left( \mathbf{x}^\prime, z^\prime \right )$,
#
# $$
# k_{\text{vel}} \left(\mathbf{X}, \mathbf{X}^{\prime}\right)=
# \begin{cases}k^{(z)}\left(\mathbf{x}, \mathbf{x}^{\prime}\right) & \text { if }
# z=z^{\prime} \\ 0 & \text { if } z \neq z^{\prime}, \end{cases}
# $$
#
# where $k^{(z)}\left(\mathbf{x}, \mathbf{x}^{\prime}\right)$ are the user chosen kernels for each dimension. What this means is that there are no correlations between the $x^{(0)}$ and $x^{(1)}$ dimensions for all choices $\mathbf{X}$ and $\mathbf{X}^{\prime}$, since there are no off-diagonal elements in the Gram matrix populated by this choice.
#
# To implement this approach in GPJax, we define `VelocityKernel` in the following cell, following the steps outlined in the [custom kernels notebook](https://docs.jaxgaussianprocesses.com/examples/constructing_new_kernels/#custom-kernel). This modular implementation takes the choice of user kernels as its class attributes: `kernel0` and `kernel1`. We must additionally pass the argument `active_dims = [0,1]`, which is an attribute of the base class `AbstractKernel`, into the chosen kernels. This is necessary such that the subsequent likelihood optimisation does not optimise over the artificial label dimension.
#


# %%


class VelocityKernel(gpx.kernels.AbstractKernel):
    def __init__(
        self,
        kernel0: gpx.kernels.AbstractKernel = gpx.kernels.RBF(active_dims=[0, 1]),
        kernel1: gpx.kernels.AbstractKernel = gpx.kernels.RBF(active_dims=[0, 1]),
    ):
        self.kernel0 = kernel0
        self.kernel1 = kernel1
        super().__init__(compute_engine=DenseKernelComputation())

    def __call__(
        self, X: Float[Array, "1 D"], Xp: Float[Array, "1 D"]
    ) -> Float[Array, "1"]:
        # standard RBF-SE kernel is x and x' are on the same output, otherwise returns 0

        z = jnp.array(X[2], dtype=int)
        zp = jnp.array(Xp[2], dtype=int)

        # achieve the correct value via 'switches' that are either 1 or 0
        k0_switch = ((z + 1) % 2) * ((zp + 1) % 2)
        k1_switch = z * zp

        return k0_switch * self.kernel0(X, Xp) + k1_switch * self.kernel1(X, Xp)


# %% [markdown]
# ### GPJax implementation
# Next, we define the model in GPJax. The prior is defined using $k_{\text{vel}}\left(\mathbf{X}, \mathbf{X}^\prime \right)$ and 0 mean and 0 observation noise. We choose a Gaussian marginal log-likelihood (MLL).
#


# %%
def initialise_gp(kernel, mean, dataset):
    prior = gpx.gps.Prior(mean_function=mean, kernel=kernel)
    likelihood = gpx.likelihoods.Gaussian(
        num_datapoints=dataset.n, obs_stddev=jnp.array([1.0e-3], dtype=jnp.float64)
    )
    posterior = prior * likelihood
    return posterior


# Define the velocity GP
mean = gpx.mean_functions.Zero()
kernel = VelocityKernel()
velocity_posterior = initialise_gp(kernel, mean, dataset_train)


# %% [markdown]
# With a model now defined, we can proceed to optimise the hyperparameters of our likelihood over $D_0$. This is done by minimising the MLL using `BFGS`. We also plot its value at each step to visually confirm that we have found the minimum. See the  [introduction to Gaussian Processes](https://docs.jaxgaussianprocesses.com/examples/intro_to_gps/) notebook for more information on optimising the MLL.


# %%
def optimise_mll(posterior, dataset, NIters=1000, key=key):
    # define the MLL using dataset_train
    objective = lambda p, d: -gpx.objectives.conjugate_mll(p, d)
    # Optimise to minimise the MLL
    opt_posterior, history = gpx.fit_scipy(
        model=posterior,
        objective=objective,
        train_data=dataset,
    )
    return opt_posterior


opt_velocity_posterior = optimise_mll(velocity_posterior, dataset_train)


# %% [markdown]
# ### Comparison
# We next obtain the latent distribution of the GP of $g$ at $\mathbf{x}_{0,i}$, then extract its mean and standard at the test locations, $\mathbf{F}_{\text{latent}}(\mathbf{x}_{0,i})$, as well as the standard deviation (we will use it at the very end).


# %%
def latent_distribution(opt_posterior, pos_3d, dataset_train):
    latent = opt_posterior.predict(pos_3d, train_data=dataset_train)
    latent_mean = latent.mean()
    latent_std = latent.stddev()
    return latent_mean, latent_std


# extract latent mean and std of g, redistribute into vectors to model F
velocity_mean, velocity_std = latent_distribution(
    opt_velocity_posterior, dataset_ground_truth.X, dataset_train
)

dataset_latent_velocity = dataset_3d(pos_test, velocity_mean)


# %% [markdown]
# We now replot the ground truth (testing data) $D_0$, the predicted latent vector field $\mathbf{F}_{\text{latent}}(\mathbf{x_i})$, and a heatmap of the residuals at each location $\mathbf{R}(\mathbf{x}_{0,i}) = \mathbf{y}_{0,i} - \mathbf{F}_{\text{latent}}(\mathbf{x}_{0,i}) $, as well as $\left|\left|\mathbf{R}(\mathbf{x}_{0,i})\right|\right|$.


# %%
# Residuals between ground truth and estimate


def plot_vector_field(ax, dataset, **kwargs):
    ax.quiver(
        dataset.X[::2][:, 0],
        dataset.X[::2][:, 1],
        dataset.y[::2],
        dataset.y[1::2],
        **kwargs,
    )


def prepare_ax(ax, X, Y, title, **kwargs):
    ax.set(
        xlim=[X.min() - 0.1, X.max() + 0.1],
        ylim=[Y.min() + 0.1, Y.max() + 0.1],
        aspect="equal",
        title=title,
        ylabel="latitude",
        **kwargs,
    )


def residuals(dataset_latent, dataset_ground_truth):
    return jnp.sqrt(
        (dataset_latent.y[::2] - dataset_ground_truth.y[::2]) ** 2
        + (dataset_latent.y[1::2] - dataset_ground_truth.y[1::2]) ** 2
    )


def plot_fields(
    dataset_ground_truth, dataset_trajectory, dataset_latent, shape=shape, scale=10
):
    X = dataset_ground_truth.X[:, 0][::2]
    Y = dataset_ground_truth.X[:, 1][::2]
    # make figure
    fig, ax = plt.subplots(1, 3, figsize=(12.0, 3.0), sharey=True)

    # ground truth
    plot_vector_field(
        ax[0],
        dataset_ground_truth,
        color=colors[0],
        label="Ocean Current",
        angles="xy",
        scale=scale,
    )
    plot_vector_field(
        ax[0],
        dataset_trajectory,
        color=colors[1],
        label="Drifter",
        angles="xy",
        scale=scale,
    )
    prepare_ax(ax[0], X, Y, "Ground Truth", xlabel="Longitude")

    # Latent estimate of vector field F
    plot_vector_field(ax[1], dataset_latent, color=colors[0], angles="xy", scale=scale)
    plot_vector_field(
        ax[1], dataset_trajectory, color=colors[1], angles="xy", scale=scale
    )
    prepare_ax(ax[1], X, Y, "GP Estimate", xlabel="Longitude")

    # residuals
    residuals_vel = jnp.flip(
        residuals(dataset_latent, dataset_ground_truth).reshape(shape), axis=0
    )
    im = ax[2].imshow(
        residuals_vel,
        extent=[X.min(), X.max(), Y.min(), Y.max()],
        cmap="jet",
        vmin=0,
        vmax=1.0,
        interpolation="spline36",
    )
    plot_vector_field(
        ax[2], dataset_trajectory, color=colors[1], angles="xy", scale=scale
    )
    prepare_ax(ax[2], X, Y, "Residuals", xlabel="Longitude")
    fig.colorbar(im, fraction=0.027, pad=0.04, orientation="vertical")

    fig.legend(
        framealpha=0.0,
        ncols=2,
        fontsize="medium",
        bbox_to_anchor=(0.5, -0.03),
        loc="lower center",
    )
    plt.show()


plot_fields(dataset_ground_truth, dataset_train, dataset_latent_velocity)


# %% [markdown]
# From the latent estimate we can see the velocity GP struggles to reconstruct features of the ground truth. This is because our construction of the kernel placed an independent prior on each physical dimension, which cannot be assumed. Therefore, we need a different approach that can implicitly incorporate this dependence at a fundamental level. To achieve this we will require a *Helmholtz Decomposition*.


# %% [markdown]
# ## Helmholtz decomposition
# In 2 dimensions, a twice continuously differentiable and compactly supported vector field $\mathbf{F}: \mathbb{R}^2 \rightarrow \mathbb{R}^2$ can be expressed as the sum of the gradient of a scalar potential $\Phi: \mathbb{R}^2 \rightarrow \mathbb{R}$, called the potential function, and the vorticity operator of another scalar potential $\Psi: \mathbb{R}^2 \rightarrow \mathbb{R}$, called the stream function ([Berlinghieri et al. (2023)](https://arxiv.org/pdf/2302.10364.pdf)) such that
# $$
# \mathbf{F}=\operatorname{grad} \Phi+\operatorname{rot} \Psi,
# $$
# where
# $$
# \operatorname{grad} \Phi:=\left[\begin{array}{l}
# \partial \Phi / \partial x^{(0)} \\
# \partial \Phi / \partial x^{(1)}
# \end{array}\right] \text { and } \operatorname{rot} \Psi:=\left[\begin{array}{c}
# \partial \Psi / \partial x^{(1)} \\
# -\partial \Psi / \partial x^{(0)}
# \end{array}\right].
# $$
#
# This is reminiscent of a 3 dimensional [Helmholtz decomposition](https://en.wikipedia.org/wiki/Helmholtz_decomposition).
#
# The 2 dimensional decomposition motivates a different approach: placing priors on $\Psi$ and $\Phi$, allowing us to make assumptions directly about fundamental properties of $\mathbf{F}$. If we choose independent GP priors such that $\Phi \sim \mathcal{G P}\left(0, k_{\Phi}\right)$ and $\Psi \sim \mathcal{G P}\left(0, k_{\Psi}\right)$, then $\mathbf{F} \sim \mathcal{G P} \left(0, k_\text{Helm}\right)$ (since acting linear operations on a GPs give GPs).
#
# For $\mathbf{X}, \mathbf{X}^{\prime} \in \mathbb{R}^2 \times \left\{0,1\right\}$ and $z, z^\prime \in \{0,1\}$,
#
# $$
# \boxed{ k_{\mathrm{Helm}}\left(\mathbf{x}, \mathbf{x}^{\prime}\right)_{z,z^\prime} =  \frac{\partial^2 k_{\Phi}\left(\mathbf{x}, \mathbf{x}^{\prime}\right)}{\partial x^{(z)} \partial\left(x^{\prime}\right)^{(z^\prime)}}+(-1)^{z+z^\prime} \frac{\partial^2 k_{\Psi}\left(\mathbf{x}, \mathbf{x}^{\prime}\right)}{\partial x^{(1-z)} \partial\left(x^{\prime}\right)^{(1-z^\prime)}}}.
# $$
#
# where $x^{(z)}$ and $(x^\prime)^{(z^\prime)}$ are the $z$ and $z^\prime$ components of $\mathbf{X}$ and ${\mathbf{X}}^{\prime}$ respectively.
#
# We compute the second derivatives using `jax.hessian`. In the following implementation, for a kernel $k(\mathbf{x}, \mathbf{x}^{\prime})$, this computes the Hessian matrix with respect to the components of $\mathbf{x}$
#
# $$
# \frac{\partial^2 k\left(\mathbf{x}, \mathbf{x}^{\prime}\right)}{\partial x^{(z)} \partial x^{(z^\prime)}}.
# $$
#
# Note that we have operated $\dfrac{\partial}{\partial x^{(z)}}$, *not* $\dfrac{\partial}{\partial \left(x^\prime \right)^{(z)}}$, as the boxed equation suggests. This is not an issue if we choose stationary kernels $k(\mathbf{x}, \mathbf{x}^{\prime}) = k(\mathbf{x} - \mathbf{x}^{\prime})$ , as the partial derivatives with respect to the components have the following exchange symmetry:
#
# $$
# \frac{\partial}{\partial x^{(z)}} = - \frac{\partial}{\partial \left( x^\prime \right)^{(z)}},
# $$
#
# for either $z$.
# %%
@dataclass
class HelmholtzKernel(gpx.kernels.stationary.StationaryKernel):
    # initialise Phi and Psi kernels as any stationary kernel in gpJax
    potential_kernel: gpx.kernels.stationary.StationaryKernel = field(
        default_factory=lambda: gpx.kernels.RBF(active_dims=[0, 1])
    )
    stream_kernel: gpx.kernels.stationary.StationaryKernel = field(
        default_factory=lambda: gpx.kernels.RBF(active_dims=[0, 1])
    )
    compute_engine = DenseKernelComputation()

    def __call__(
        self, X: Float[Array, "1 D"], Xp: Float[Array, "1 D"]
    ) -> Float[Array, "1"]:
        # obtain indices for k_helm, implement in the correct sign between the derivatives
        z = jnp.array(X[2], dtype=int)
        zp = jnp.array(Xp[2], dtype=int)
        sign = (-1) ** (z + zp)

        # convert to array to correctly index, -ve sign due to exchange symmetry (only true for stationary kernels)
        potential_dvtve = -jnp.array(
            hessian(self.potential_kernel)(X, Xp), dtype=jnp.float64
        )[z][zp]
        stream_dvtve = -jnp.array(
            hessian(self.stream_kernel)(X, Xp), dtype=jnp.float64
        )[1 - z][1 - zp]

        return potential_dvtve + sign * stream_dvtve


# %% [markdown]
# ### GPJax implementation
# We repeat the same steps as with the velocity GP model, replacing `VelocityKernel` with `HelmholtzKernel`.

# %%
# Redefine Gaussian process with Helmholtz kernel
kernel = HelmholtzKernel()
helmholtz_posterior = initialise_gp(kernel, mean, dataset_train)
# Optimise hyperparameters using BFGS
opt_helmholtz_posterior = optimise_mll(helmholtz_posterior, dataset_train)


# %% [markdown]
# ### Comparison
# We again plot the ground truth (testing data) $D_0$, the predicted latent vector field $\mathbf{F}_{\text{latent}}(\mathbf{x}_{0,i})$, and a heatmap of the residuals at each location $R(\mathbf{x}_{0,i}) = \mathbf{y}_{0,i} - \mathbf{F}_{\text{latent}}(\mathbf{x}_{0,i})$ and $\left|\left|R(\mathbf{x}_{0,i})  \right|\right|$.

# %%
# obtain latent distribution, extract x and y values over g
helmholtz_mean, helmholtz_std = latent_distribution(
    opt_helmholtz_posterior, dataset_ground_truth.X, dataset_train
)
dataset_latent_helmholtz = dataset_3d(pos_test, helmholtz_mean)

plot_fields(dataset_ground_truth, dataset_train, dataset_latent_helmholtz)

# %% [markdown]
# Visually, the Helmholtz model performs better than the velocity model, preserving the local structure of the $\mathbf{F}$. Since we placed priors on $\Phi$ and $\Psi$, the construction of $\mathbf{F}$ allows for correlations between the dimensions (non-zero off-diagonal elements in the Gram matrix populated by $k_\text{Helm}\left(\mathbf{X},\mathbf{X}^{\prime}\right)$ ).


# %% [markdown]
# ## Negative log predictive densities
# Lastly, we directly compare the velocity and Helmholtz models by computing the [negative log predictive densities](https://en.wikipedia.org/wiki/Negative_log_predictive_density) for each model. This is a quantitative metric that measures the probability of the ground truth given the data.
#
# $$
# \mathrm{NLPD}=-\sum_{i=1}^{2N} \log \left(  p\left(\mathcal{Y}_i = Y_{0,i} \mid \mathbf{X}_{i}\right) \right),
# $$
#
# where each $p\left(\mathcal{Y}_i \mid \mathbf{X}_i \right)$ is the marginal Gaussian distribution over $\mathcal{Y}_i$ at each test location, and $Y_{i,0}$ is the $i$-th component of the (massaged) test data that we reserved at the beginning of the notebook in $D_0$. A smaller value is better, since the deviation of the ground truth and the model are small in this case.


# %%
# ensure testing data alternates between x0 and x1 components
def nlpd(mean, std, vel_test):
    vel_query = jnp.column_stack((vel_test[0], vel_test[1])).flatten()
    normal = tfp.substrates.jax.distributions.Normal(loc=mean, scale=std)
    return -jnp.sum(normal.log_prob(vel_query))


# compute nlpd for velocity and helmholtz
nlpd_vel = nlpd(velocity_mean, velocity_std, vel_test)
nlpd_helm = nlpd(helmholtz_mean, helmholtz_std, vel_test)

print("NLPD for Velocity: %.2f \nNLPD for Helmholtz: %.2f" % (nlpd_vel, nlpd_helm))
# %% [markdown]
# The Helmholtz model outperforms the velocity model, as indicated by the lower NLPD score.

# %% [markdown]
# <span id="fn1"></span>
# ## Footnote
# Kernels for vector-valued functions have been studied in the literature, see [Alvarez et al. (2012)](https://doi.org/10.48550/arXiv.1106.6251)
# ## System configuration
# %%
# %reload_ext watermark
# %watermark -n -u -v -iv -w -a 'Ivan Shalashilin'



# Copyright 2024 The GPJax Contributors. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ==============================================================================
from jax import config

config.update("jax_enable_x64", True)

import jax
import jax.numpy as jnp
import jax.random as jr

from gpjax.decision_making.test_functions.continuous_functions import Forrester
from gpjax.decision_making.utility_functions.probability_of_improvement import (
    ProbabilityOfImprovement,
)
from gpjax.decision_making.utils import OBJECTIVE
from tests.test_decision_making.utils import generate_dummy_conjugate_posterior


def test_probability_of_improvement_gives_correct_value_for_a_seed():
    key = jr.key(42)
    forrester = Forrester()
    dataset = forrester.generate_dataset(num_points=10, key=key)
    posterior = generate_dummy_conjugate_posterior(dataset, forrester)
    posteriors = {OBJECTIVE: posterior}
    datasets = {OBJECTIVE: dataset}

    pi_utility_builder = ProbabilityOfImprovement()
    pi_utility = pi_utility_builder.build_utility_function(
        posteriors=posteriors, datasets=datasets, key=key
    )

    test_X = forrester.generate_test_points(num_points=10, key=key)
    utility_values = pi_utility(test_X)

    # Computing the expected utility values
    predictive_dist = posterior.predict(test_X, train_data=dataset)
    predictive_mean = predictive_dist.mean()
    predictive_std = predictive_dist.stddev()

    # Computing best_y as the min. of the posterior predictive mean
    # over the training set.
    predictive_dist_for_training_data = posterior.predict(dataset.X, train_data=dataset)
    best_y = predictive_dist_for_training_data.mean().min()

    # Gaussian CDF computed "by hand"
    x_ = (best_y - predictive_mean) / predictive_std
    expected_utility_values = 0.5 * (
        1 + jax.scipy.special.erf(x_ / jnp.sqrt(2))
    ).reshape(-1, 1)

    assert utility_values.shape == (10, 1)
    assert jnp.isclose(utility_values, expected_utility_values).all()



# Copyright 2023 The JaxGaussianProcesses Contributors. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ==============================================================================

from flax import nnx
import jax.numpy as jnp
import jax.random as jr
from jaxtyping import (
    Float,
    Num,
)
import optax as ox
import pytest
import scipy

from gpjax.dataset import Dataset
from gpjax.fit import (
    fit,
    fit_scipy,
    get_batch,
)
from gpjax.gps import (
    ConjugatePosterior,
    Prior,
)
from gpjax.kernels import RBF
from gpjax.likelihoods import Gaussian
from gpjax.mean_functions import (
    AbstractMeanFunction,
    Constant,
)
from gpjax.objectives import (
    conjugate_mll,
    elbo,
)
from gpjax.parameters import (
    PositiveReal,
    Static,
)
from gpjax.typing import Array
from gpjax.variational_families import VariationalGaussian


def test_fit_simple() -> None:
    # Create dataset:
    X = jnp.linspace(0.0, 10.0, 100).reshape(-1, 1)
    y = 2.0 * X + 1.0 + 10 * jr.normal(jr.key(0), X.shape).reshape(-1, 1)
    D = Dataset(X, y)

    # Define linear model:

    class LinearModel(nnx.Module):
        def __init__(self, weight: float, bias: float):
            self.weight = PositiveReal(weight)
            self.bias = Static(bias)

        def __call__(self, x):
            return self.weight.value * x + self.bias.value

    model = LinearModel(weight=1.0, bias=1.0)

    # Define loss function:
    def mse(model, data):
        pred = model(data.X)
        return jnp.mean((pred - data.y) ** 2)

    # Train!
    trained_model, hist = fit(
        model=model,
        objective=mse,
        train_data=D,
        optim=ox.sgd(0.001),
        num_iters=100,
        key=jr.key(123),
    )

    # Ensure we return a history of the correct length
    assert len(hist) == 100

    # Ensure we return a model of the same class
    assert isinstance(trained_model, LinearModel)

    # Test reduction in loss:
    assert mse(trained_model, D) < mse(model, D)

    # Test stop_gradient on bias:
    assert trained_model.bias.value == 1.0


def test_fit_scipy_simple():
    # Create dataset:
    X = jnp.linspace(0.0, 10.0, 100).reshape(-1, 1)
    y = 2.0 * X + 1.0 + 10 * jr.normal(jr.PRNGKey(0), X.shape).reshape(-1, 1)
    D = Dataset(X, y)

    # Define linear model:
    class LinearModel(nnx.Module):
        def __init__(self, weight: float, bias: float):
            self.weight = PositiveReal(weight)
            self.bias = Static(bias)

        def __call__(self, x):
            return self.weight.value * x + self.bias.value

    model = LinearModel(weight=1.0, bias=1.0)

    # Define loss function:
    def mse(model, data):
        pred = model(data.X)
        return jnp.mean((pred - data.y) ** 2)

    # Train with bfgs!
    trained_model, hist = fit_scipy(
        model=model,
        objective=mse,
        train_data=D,
        max_iters=10,
    )

    # Ensure we return a history of the correct length
    assert len(hist) > 2

    # Ensure we return a model of the same class
    assert isinstance(trained_model, LinearModel)

    # Test reduction in loss:
    assert mse(trained_model, D) < mse(model, D)

    # Test stop_gradient on bias:
    assert trained_model.bias.value == 1.0


@pytest.mark.parametrize("n_data", [20])
@pytest.mark.parametrize("verbose", [True, False])
def test_fit_gp_regression(n_data: int, verbose: bool) -> None:
    # Create dataset:
    key = jr.PRNGKey(123)
    x = jnp.sort(
        jr.uniform(key=key, minval=-2.0, maxval=2.0, shape=(n_data, 1)), axis=0
    )
    y = jnp.sin(x) + jr.normal(key=key, shape=x.shape) * 0.1
    D = Dataset(X=x, y=y)

    # Define GP model:
    prior = Prior(kernel=RBF(), mean_function=Constant())
    likelihood = Gaussian(num_datapoints=n_data)
    posterior = prior * likelihood

    # Train!
    trained_model, history = fit(
        model=posterior,
        objective=conjugate_mll,
        train_data=D,
        optim=ox.adam(0.1),
        num_iters=15,
        verbose=verbose,
        key=jr.PRNGKey(123),
    )

    # Ensure the trained model is a Gaussian process posterior
    assert isinstance(trained_model, ConjugatePosterior)

    # Ensure we return a history of the correct length
    assert len(history) == 15

    # Ensure we reduce the loss
    assert conjugate_mll(trained_model, D) < conjugate_mll(posterior, D)


@pytest.mark.parametrize("n_data", [20])
@pytest.mark.parametrize("verbose", [True, False])
def test_fit_scipy_gp_regression(n_data: int, verbose: bool) -> None:
    # Create dataset:
    key = jr.PRNGKey(123)
    x = jnp.sort(
        jr.uniform(key=key, minval=-2.0, maxval=2.0, shape=(n_data, 1)), axis=0
    )
    y = jnp.sin(x) + jr.normal(key=key, shape=x.shape) * 0.1
    D = Dataset(X=x, y=y)

    # Define GP model:
    prior = Prior(kernel=RBF(), mean_function=Constant())
    likelihood = Gaussian(num_datapoints=n_data)
    posterior = prior * likelihood

    # Train with BFGS!
    trained_model_bfgs, history_bfgs = fit_scipy(
        model=posterior,
        objective=conjugate_mll,
        train_data=D,
        max_iters=40,
        verbose=verbose,
    )

    # Ensure the trained model is a Gaussian process posterior
    assert isinstance(trained_model_bfgs, ConjugatePosterior)

    # Ensure we return a history_bfgs of the correct length
    assert len(history_bfgs) > 2

    # Ensure we reduce the loss
    assert conjugate_mll(trained_model_bfgs, D) < conjugate_mll(posterior, D)


def test_fit_scipy_error_raises() -> None:
    # Create dataset:
    D = Dataset(
        X=jnp.array([[0.0]], dtype=jnp.float64), y=jnp.array([[0.0]], dtype=jnp.float64)
    )

    # build crazy mean function so that opt fails
    class CrazyMean(AbstractMeanFunction):
        def __call__(self, x: Num[Array, "N D"]) -> Float[Array, "N O"]:
            return jnp.heaviside(x, 100.0)

    # Define GP model with crazy mean function:
    prior = Prior(kernel=RBF(), mean_function=CrazyMean())
    likelihood = Gaussian(num_datapoints=2)
    posterior = prior * likelihood

    with pytest.raises(scipy.optimize.OptimizeWarning):
        fit_scipy(
            model=posterior,
            objective=conjugate_mll,
            train_data=D,
            max_iters=10,
        )

    # also check fails if no given enough steps
    prior = Prior(kernel=RBF(), mean_function=Constant())
    likelihood = Gaussian(num_datapoints=2)
    posterior = prior * likelihood

    with pytest.raises(scipy.optimize.OptimizeWarning):
        fit_scipy(
            model=posterior,
            objective=conjugate_mll,
            train_data=D,
            max_iters=1,
        )


@pytest.mark.parametrize("num_iters", [1, 5])
@pytest.mark.parametrize("batch_size", [1, 20, 50])
@pytest.mark.parametrize("n_data", [50])
@pytest.mark.parametrize("verbose", [True, False])
def test_fit_batch(num_iters: int, batch_size: int, n_data: int, verbose: bool) -> None:
    # Create dataset:
    key = jr.key(123)
    x = jnp.sort(
        jr.uniform(key=key, minval=-2.0, maxval=2.0, shape=(n_data, 1)), axis=0
    )
    y = jnp.sin(x) + jr.normal(key=key, shape=x.shape) * 0.1
    D = Dataset(X=x, y=y)

    # Define GP model:
    prior = Prior(kernel=RBF(), mean_function=Constant())
    likelihood = Gaussian(num_datapoints=n_data)
    posterior = prior * likelihood

    # Define variational family:
    z = jnp.linspace(-2.0, 2.0, 10).reshape(-1, 1)
    q = VariationalGaussian(posterior=posterior, inducing_inputs=z)

    # Train!
    trained_model, history = fit(
        model=q,
        objective=elbo,
        train_data=D,
        optim=ox.adam(0.1),
        num_iters=num_iters,
        batch_size=batch_size,
        verbose=verbose,
        key=jr.key(123),
    )

    # Ensure the trained model is a Gaussian process posterior
    assert isinstance(trained_model, VariationalGaussian)

    # Ensure we return a history of the correct length
    assert len(history) == num_iters

    # Ensure we reduce the loss
    assert elbo(trained_model, D) < elbo(q, D)


@pytest.mark.parametrize("n_data", [50])
@pytest.mark.parametrize("n_dim", [1, 2, 3])
@pytest.mark.parametrize("batch_size", [1, 2, 50])
def test_get_batch(n_data: int, n_dim: int, batch_size: int):
    key = jr.key(123)

    # Create dataset:
    x = jnp.sort(
        jr.uniform(key=key, minval=-2.0, maxval=2.0, shape=(n_data, n_dim)), axis=0
    )
    y = jnp.sin(x) + jr.normal(key=key, shape=x.shape) * 0.1
    D = Dataset(X=x, y=y)

    # Sample out a batch:
    B = get_batch(D, batch_size, key)

    # Check batch is correct size and shape dimensions:
    assert B.n == batch_size
    assert B.X.shape[1:] == x.shape[1:]
    assert B.y.shape[1:] == y.shape[1:]

    # Ensure no caching of batches:
    key, subkey = jr.split(key)
    New = get_batch(D, batch_size, subkey)
    assert New.n == batch_size
    assert New.X.shape[1:] == x.shape[1:]
    assert New.y.shape[1:] == y.shape[1:]
    assert (New.X != B.X).all()
    assert (New.y != B.y).all()



# Copyright 2022 The JaxGaussianProcesses Contributors. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ==============================================================================

from gpjax.kernels.stationary.base import StationaryKernel
from gpjax.kernels.stationary.matern12 import Matern12
from gpjax.kernels.stationary.matern32 import Matern32
from gpjax.kernels.stationary.matern52 import Matern52
from gpjax.kernels.stationary.periodic import Periodic
from gpjax.kernels.stationary.powered_exponential import PoweredExponential
from gpjax.kernels.stationary.rational_quadratic import RationalQuadratic
from gpjax.kernels.stationary.rbf import RBF
from gpjax.kernels.stationary.white import White

__all__ = [
    "Matern12",
    "Matern32",
    "Matern52",
    "Periodic",
    "PoweredExponential",
    "RationalQuadratic",
    "RBF",
    "StationaryKernel",
    "White",
]



# Copyright 2023 The JaxGaussianProcesses Contributors. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ==============================================================================
from jax import config

config.update("jax_enable_x64", True)

import jax.numpy as jnp
import jax.random as jr
import optax as ox
import pytest

import gpjax as gpx
from gpjax.dataset import Dataset
from gpjax.decision_making.decision_maker import (
    AbstractDecisionMaker,
    UtilityDrivenDecisionMaker,
)
from gpjax.decision_making.posterior_handler import PosteriorHandler
from gpjax.decision_making.search_space import (
    AbstractSearchSpace,
    ContinuousSearchSpace,
)
from gpjax.decision_making.test_functions import Quadratic
from gpjax.decision_making.utility_functions import (
    AbstractSinglePointUtilityFunctionBuilder,
    ThompsonSampling,
)
from gpjax.decision_making.utility_maximizer import (
    AbstractSinglePointUtilityMaximizer,
    ContinuousSinglePointUtilityMaximizer,
)
from gpjax.decision_making.utils import (
    OBJECTIVE,
    build_function_evaluator,
)
from gpjax.typing import KeyArray
from tests.test_decision_making.utils import QuadraticSinglePointUtilityFunctionBuilder

CONSTRAINT = "CONSTRAINT"


@pytest.fixture
def search_space() -> ContinuousSearchSpace:
    return ContinuousSearchSpace(
        lower_bounds=jnp.array([0.0], dtype=jnp.float64),
        upper_bounds=jnp.array([1.0], dtype=jnp.float64),
    )


@pytest.fixture
def posterior_handler() -> PosteriorHandler:
    mean = gpx.mean_functions.Zero()
    kernel = gpx.kernels.Matern52(
        lengthscale=jnp.array(1.0),
        variance=jnp.array(1.0),
        n_dims=1,
    )
    prior = gpx.gps.Prior(mean_function=mean, kernel=kernel)
    likelihood_builder = lambda x: gpx.likelihoods.Gaussian(
        num_datapoints=x, obs_stddev=jnp.array(1e-3)
    )
    posterior_handler = PosteriorHandler(
        prior=prior,
        likelihood_builder=likelihood_builder,
        optimization_objective=gpx.objectives.conjugate_mll,
        optimizer=ox.adam(learning_rate=0.01),
        num_optimization_iters=100,
    )
    return posterior_handler


@pytest.fixture
def utility_function_builder() -> AbstractSinglePointUtilityFunctionBuilder:
    return QuadraticSinglePointUtilityFunctionBuilder()


@pytest.fixture
def thompson_sampling_utility_function_builder() -> ThompsonSampling:
    return ThompsonSampling(num_features=100)


@pytest.fixture
def utility_maximizer() -> AbstractSinglePointUtilityMaximizer:
    return ContinuousSinglePointUtilityMaximizer(
        num_initial_samples=1000, num_restarts=1
    )


def get_dataset(num_points: int, key: KeyArray) -> Dataset:
    test_function = Quadratic()
    dataset = test_function.generate_dataset(num_points=num_points, key=key)
    return dataset


def test_abstract_decision_maker_raises_error():
    with pytest.raises(TypeError):
        AbstractDecisionMaker()


@pytest.mark.parametrize("batch_size", [0, -1, -10])
def test_invalid_batch_size_raises_error(
    search_space: AbstractSearchSpace,
    posterior_handler: PosteriorHandler,
    utility_function_builder: AbstractSinglePointUtilityFunctionBuilder,
    utility_maximizer: AbstractSinglePointUtilityMaximizer,
    batch_size: int,
):
    key = jr.key(42)
    posterior_handlers = {OBJECTIVE: posterior_handler}
    objective_dataset = get_dataset(num_points=5, key=jr.key(42))
    datasets = {"OBJECTIVE": objective_dataset}
    with pytest.raises(ValueError):
        UtilityDrivenDecisionMaker(
            search_space=search_space,
            posterior_handlers=posterior_handlers,
            datasets=datasets,
            utility_function_builder=utility_function_builder,
            utility_maximizer=utility_maximizer,
            key=key,
            post_ask=[],
            post_tell=[],
            batch_size=batch_size,
        )


def test_non_thompson_sampling_non_one_batch_size_raises_error(
    search_space: AbstractSearchSpace,
    posterior_handler: PosteriorHandler,
    utility_function_builder: AbstractSinglePointUtilityFunctionBuilder,
    utility_maximizer: AbstractSinglePointUtilityMaximizer,
):
    key = jr.key(42)
    posterior_handlers = {OBJECTIVE: posterior_handler}
    objective_dataset = get_dataset(num_points=5, key=jr.key(42))
    datasets = {"OBJECTIVE": objective_dataset}
    with pytest.raises(NotImplementedError):
        UtilityDrivenDecisionMaker(
            search_space=search_space,
            posterior_handlers=posterior_handlers,
            datasets=datasets,
            utility_function_builder=utility_function_builder,
            utility_maximizer=utility_maximizer,
            key=key,
            post_ask=[],
            post_tell=[],
            batch_size=2,
        )


def test_invalid_tags_raises_error(
    search_space: AbstractSearchSpace,
    posterior_handler: PosteriorHandler,
    utility_function_builder: AbstractSinglePointUtilityFunctionBuilder,
    utility_maximizer: AbstractSinglePointUtilityMaximizer,
):
    key = jr.key(42)
    posterior_handlers = {OBJECTIVE: posterior_handler}
    dataset = get_dataset(num_points=5, key=jr.key(42))
    datasets = {"CONSTRAINT": dataset}  # Dataset tag doesn't match posterior tag
    with pytest.raises(ValueError):
        UtilityDrivenDecisionMaker(
            search_space=search_space,
            posterior_handlers=posterior_handlers,
            datasets=datasets,
            utility_function_builder=utility_function_builder,
            utility_maximizer=utility_maximizer,
            key=key,
            post_ask=[],
            post_tell=[],
            batch_size=1,
        )


def test_initialisation_optimizes_posterior_hyperparameters(
    search_space: AbstractSearchSpace,
    posterior_handler: PosteriorHandler,
    utility_function_builder: AbstractSinglePointUtilityFunctionBuilder,
    utility_maximizer: AbstractSinglePointUtilityMaximizer,
):
    key = jr.key(42)
    posterior_handlers = {OBJECTIVE: posterior_handler, CONSTRAINT: posterior_handler}
    objective_dataset = get_dataset(num_points=5, key=jr.key(42))
    constraint_dataset = get_dataset(num_points=5, key=jr.key(10))
    datasets = {"OBJECTIVE": objective_dataset, CONSTRAINT: constraint_dataset}
    decision_maker = UtilityDrivenDecisionMaker(
        search_space=search_space,
        posterior_handlers=posterior_handlers,
        datasets=datasets,
        utility_function_builder=utility_function_builder,
        utility_maximizer=utility_maximizer,
        key=key,
        post_ask=[],
        post_tell=[],
        batch_size=1,
    )
    # Assert kernel hyperparameters get changed from their initial values
    assert decision_maker.posteriors[OBJECTIVE].prior.kernel.lengthscale != jnp.array(
        1.0
    )
    assert decision_maker.posteriors[OBJECTIVE].prior.kernel.variance != jnp.array(1.0)
    assert decision_maker.posteriors[CONSTRAINT].prior.kernel.lengthscale != jnp.array(
        1.0
    )
    assert decision_maker.posteriors[CONSTRAINT].prior.kernel.variance != jnp.array(1.0)
    assert (
        decision_maker.posteriors[CONSTRAINT].prior.kernel.lengthscale
        != decision_maker.posteriors[OBJECTIVE].prior.kernel.lengthscale
    )
    assert (
        decision_maker.posteriors[CONSTRAINT].prior.kernel.variance
        != decision_maker.posteriors[OBJECTIVE].prior.kernel.variance
    )


def test_decision_maker_ask(
    search_space: AbstractSearchSpace,
    posterior_handler: PosteriorHandler,
    utility_function_builder: AbstractSinglePointUtilityFunctionBuilder,
    utility_maximizer: AbstractSinglePointUtilityMaximizer,
):
    key = jr.key(42)
    posterior_handlers = {OBJECTIVE: posterior_handler}
    objective_dataset = get_dataset(num_points=5, key=jr.key(42))
    datasets = {"OBJECTIVE": objective_dataset}
    decision_maker = UtilityDrivenDecisionMaker(
        search_space=search_space,
        posterior_handlers=posterior_handlers,
        datasets=datasets,
        utility_function_builder=utility_function_builder,
        utility_maximizer=utility_maximizer,
        key=key,
        post_ask=[],
        post_tell=[],
        batch_size=1,
    )
    initial_decision_maker_key = decision_maker.key
    query_point = decision_maker.ask(key=key)
    assert query_point.shape == (1, 1)
    assert jnp.allclose(query_point, jnp.array([[0.5]]), atol=1e-5)
    assert len(decision_maker.current_utility_functions) == 1
    assert (
        decision_maker.key == initial_decision_maker_key
    ).all()  # Ensure decision maker key is unchanged


@pytest.mark.parametrize("batch_size", [1, 2, 5])
def test_decision_maker_ask_multi_batch_ts(
    search_space: AbstractSearchSpace,
    posterior_handler: PosteriorHandler,
    thompson_sampling_utility_function_builder: ThompsonSampling,
    utility_maximizer: AbstractSinglePointUtilityMaximizer,
    batch_size: int,
):
    key = jr.key(42)
    posterior_handlers = {OBJECTIVE: posterior_handler}
    objective_dataset = get_dataset(num_points=5, key=jr.key(42))
    datasets = {"OBJECTIVE": objective_dataset}
    decision_maker = UtilityDrivenDecisionMaker(
        search_space=search_space,
        posterior_handlers=posterior_handlers,
        datasets=datasets,
        utility_function_builder=thompson_sampling_utility_function_builder,
        utility_maximizer=utility_maximizer,
        key=key,
        post_ask=[],
        post_tell=[],
        batch_size=batch_size,
    )
    initial_decision_maker_key = decision_maker.key
    query_points = decision_maker.ask(key=key)
    assert query_points.shape == (batch_size, 1)

    # TODO: ask henry about this failing assertion
    # assert (
    #     len(jnp.unique(query_points)) == batch_size
    # )  # Ensure we aren't drawing the same Thompson sample each time
    assert len(decision_maker.current_utility_functions) == batch_size
    assert (
        decision_maker.key == initial_decision_maker_key
    ).all()  # Ensure decision maker key is unchanged


def test_decision_maker_tell_with_inconsistent_observations_raises_error(
    search_space: AbstractSearchSpace,
    posterior_handler: PosteriorHandler,
    utility_function_builder: AbstractSinglePointUtilityFunctionBuilder,
    utility_maximizer: AbstractSinglePointUtilityMaximizer,
):
    key = jr.key(42)
    posterior_handlers = {OBJECTIVE: posterior_handler, CONSTRAINT: posterior_handler}
    initial_objective_dataset = get_dataset(num_points=5, key=jr.key(42))
    initial_constraint_dataset = get_dataset(num_points=5, key=jr.key(10))
    datasets = {
        OBJECTIVE: initial_objective_dataset,
        CONSTRAINT: initial_constraint_dataset,
    }
    decision_maker = UtilityDrivenDecisionMaker(
        search_space=search_space,
        posterior_handlers=posterior_handlers,
        datasets=datasets,
        utility_function_builder=utility_function_builder,
        utility_maximizer=utility_maximizer,
        key=key,
        post_ask=[],
        post_tell=[],
        batch_size=1,
    )
    mock_objective_observation = get_dataset(num_points=1, key=jr.key(1))
    mock_constraint_observation = get_dataset(num_points=1, key=jr.key(2))
    observations = {
        OBJECTIVE: mock_objective_observation,
        "CONSTRAINT_ONE": mock_constraint_observation,  # Deliberately incorrect tag
    }
    with pytest.raises(ValueError):
        decision_maker.tell(observation_datasets=observations, key=key)


def test_decision_maker_tell_updates_datasets_and_models(
    search_space: AbstractSearchSpace,
    posterior_handler: PosteriorHandler,
    utility_function_builder: AbstractSinglePointUtilityFunctionBuilder,
    utility_maximizer: AbstractSinglePointUtilityMaximizer,
):
    key = jr.key(42)
    posterior_handlers = {OBJECTIVE: posterior_handler, CONSTRAINT: posterior_handler}
    initial_objective_dataset = get_dataset(num_points=5, key=jr.key(42))
    initial_constraint_dataset = get_dataset(num_points=5, key=jr.key(10))
    datasets = {
        "OBJECTIVE": initial_objective_dataset,
        CONSTRAINT: initial_constraint_dataset,
    }
    decision_maker = UtilityDrivenDecisionMaker(
        search_space=search_space,
        posterior_handlers=posterior_handlers,
        datasets=datasets,
        utility_function_builder=utility_function_builder,
        utility_maximizer=utility_maximizer,
        key=key,
        post_ask=[],
        post_tell=[],
        batch_size=1,
    )
    initial_decision_maker_key = decision_maker.key
    initial_objective_posterior = decision_maker.posteriors[OBJECTIVE]
    initial_constraint_posterior = decision_maker.posteriors[CONSTRAINT]
    mock_objective_observation = get_dataset(num_points=1, key=jr.key(1))
    mock_constraint_observation = get_dataset(num_points=1, key=jr.key(2))
    observations = {
        OBJECTIVE: mock_objective_observation,
        CONSTRAINT: mock_constraint_observation,
    }
    decision_maker.tell(observation_datasets=observations, key=key)
    assert decision_maker.datasets[OBJECTIVE].n == 6
    assert decision_maker.datasets[CONSTRAINT].n == 6
    assert decision_maker.datasets[OBJECTIVE].X[-1] == mock_objective_observation.X[0]
    assert decision_maker.datasets[CONSTRAINT].X[-1] == mock_constraint_observation.X[0]
    assert (
        decision_maker.posteriors[OBJECTIVE].prior.kernel.lengthscale
        != initial_objective_posterior.prior.kernel.lengthscale
    )
    assert (
        decision_maker.posteriors[OBJECTIVE].prior.kernel.variance
        != initial_objective_posterior.prior.kernel.variance
    )
    assert (
        decision_maker.posteriors[CONSTRAINT].prior.kernel.lengthscale
        != initial_constraint_posterior.prior.kernel.lengthscale
    )
    assert (
        decision_maker.posteriors[CONSTRAINT].prior.kernel.variance
        != initial_constraint_posterior.prior.kernel.variance
    )
    assert (
        decision_maker.key == initial_decision_maker_key
    ).all()  # Ensure decision maker key has not been updated


@pytest.mark.parametrize("n_steps", [1, 3])
def test_decision_maker_run(
    search_space: AbstractSearchSpace,
    posterior_handler: PosteriorHandler,
    utility_function_builder: AbstractSinglePointUtilityFunctionBuilder,
    utility_maximizer: AbstractSinglePointUtilityMaximizer,
    n_steps: int,
):
    key = jr.key(42)
    posterior_handlers = {OBJECTIVE: posterior_handler}
    initial_objective_dataset = get_dataset(num_points=5, key=jr.key(42))
    initial_datasets = {
        "OBJECTIVE": initial_objective_dataset,
    }
    decision_maker = UtilityDrivenDecisionMaker(
        search_space=search_space,
        posterior_handlers=posterior_handlers,
        datasets=initial_datasets,
        utility_function_builder=utility_function_builder,
        utility_maximizer=utility_maximizer,
        key=key,
        post_ask=[],
        post_tell=[],
        batch_size=1,
    )
    initial_decision_maker_key = decision_maker.key
    black_box_fn = Quadratic()
    black_box_function_evaluator = build_function_evaluator(
        {OBJECTIVE: black_box_fn.evaluate}
    )
    query_datasets = decision_maker.run(
        n_steps=n_steps, black_box_function_evaluator=black_box_function_evaluator
    )
    assert initial_datasets[OBJECTIVE].n == 5
    assert query_datasets[OBJECTIVE].n == 5 + n_steps
    assert (
        jnp.abs(query_datasets[OBJECTIVE].X[-n_steps:] - jnp.array([[0.5]])) < 1e-5
    ).all()  # Ensure we're querying the correct point in our dummy utility function at each step
    assert (
        decision_maker.key != initial_decision_maker_key
    ).all()  # Ensure decision maker key gets updated


@pytest.mark.parametrize("n_steps", [1, 3])
@pytest.mark.parametrize("batch_size", [1, 3])
def test_decision_maker_run_ts(
    search_space: AbstractSearchSpace,
    posterior_handler: PosteriorHandler,
    thompson_sampling_utility_function_builder: ThompsonSampling,
    utility_maximizer: AbstractSinglePointUtilityMaximizer,
    n_steps: int,
    batch_size: int,
):
    key = jr.key(42)
    posterior_handlers = {OBJECTIVE: posterior_handler}
    initial_objective_dataset = get_dataset(num_points=5, key=jr.key(42))
    initial_datasets = {
        "OBJECTIVE": initial_objective_dataset,
    }
    decision_maker = UtilityDrivenDecisionMaker(
        search_space=search_space,
        posterior_handlers=posterior_handlers,
        datasets=initial_datasets,
        utility_function_builder=thompson_sampling_utility_function_builder,
        utility_maximizer=utility_maximizer,
        key=key,
        post_ask=[],
        post_tell=[],
        batch_size=batch_size,
    )
    initial_decision_maker_key = decision_maker.key
    black_box_fn = Quadratic()
    black_box_function_evaluator = build_function_evaluator(
        {OBJECTIVE: black_box_fn.evaluate}
    )
    query_datasets = decision_maker.run(
        n_steps=n_steps, black_box_function_evaluator=black_box_function_evaluator
    )
    assert initial_datasets[OBJECTIVE].n == 5
    assert (
        query_datasets[OBJECTIVE].n == 5 + n_steps * batch_size
    )  # Ensure we're getting the correct number of points
    assert (
        decision_maker.key != initial_decision_maker_key
    ).all()  # Ensure decision maker key gets updated



# Copyright 2024 The JaxGaussianProcesses Contributors. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ==============================================================================
from dataclasses import dataclass
from functools import partial

from beartype.typing import Mapping
import jax.numpy as jnp
import tensorflow_probability.substrates.jax as tfp

from gpjax.dataset import Dataset
from gpjax.decision_making.utility_functions.base import (
    AbstractSinglePointUtilityFunctionBuilder,
    SinglePointUtilityFunction,
)
from gpjax.decision_making.utils import (
    OBJECTIVE,
    get_best_latent_observation_val,
)
from gpjax.gps import ConjugatePosterior
from gpjax.typing import (
    Array,
    Float,
    KeyArray,
)


@dataclass
class ExpectedImprovement(AbstractSinglePointUtilityFunctionBuilder):
    """
    Expected Improvement acquisition function as introduced by [Močkus,
    1974](https://link.springer.com/chapter/10.1007/3-540-07165-2_55). The "best"
    incumbent value is defined as the lowest posterior mean value evaluated at the the
    previously observed points. This enables the acquisition function to be utilised with noisy observations.
    """

    def build_utility_function(
        self,
        posteriors: Mapping[str, ConjugatePosterior],
        datasets: Mapping[str, Dataset],
        key: KeyArray,
    ) -> SinglePointUtilityFunction:
        r"""
        Build the Expected Improvement acquisition function. This computes the expected
        improvement over the "best" of the previously observed points, utilising the
        posterior distribution of the surrogate model. For posterior distribution
        $`f(\cdot)`$, and best incumbent value $`\eta`$, this is defined
        as:
        ```math
        \alpha_{\text{EI}}(\mathbf{x}) = \mathbb{E}\left[\max(0, \eta - f(\mathbf{x}))\right]
        ```

        Args:
            posteriors (Mapping[str, ConjugatePosterior]): Dictionary of posteriors to
            used to form the utility function. One posteriors must correspond to the
            `OBJECTIVE` key, as we utilise the objective posterior to form the utility
            function.
            datasets (Mapping[str, Dataset]): Dictionary of datasets used to form the
            utility function. Keys in `datasets` should correspond to keys in
            `posteriors`. One of the datasets must correspond to the `OBJECTIVE` key.
            key (KeyArray): JAX PRNG key used for random number generation.

        Returns:
            SinglePointUtilityFunction: The Expected Improvement acquisition function to
            to be *maximised* in order to decide which point to query next.
        """
        self.check_objective_present(posteriors, datasets)
        objective_posterior = posteriors[OBJECTIVE]
        objective_dataset = datasets[OBJECTIVE]

        if not isinstance(objective_posterior, ConjugatePosterior):
            raise ValueError(
                "Objective posterior must be a ConjugatePosterior to compute the Expected Improvement."
            )

        if (
            objective_dataset.X is None
            or objective_dataset.n == 0
            or objective_dataset.y is None
        ):
            raise ValueError("Objective dataset must contain at least one item")

        eta = get_best_latent_observation_val(objective_posterior, objective_dataset)
        return partial(
            _expected_improvement, objective_posterior, objective_dataset, eta
        )


def _expected_improvement(
    objective_posterior: ConjugatePosterior,
    objective_dataset: Dataset,
    eta: Float[Array, ""],
    x: Float[Array, "N D"],
) -> Float[Array, "N 1"]:
    latent_dist = objective_posterior(x, objective_dataset)
    mean = latent_dist.mean()
    var = latent_dist.variance()
    normal = tfp.distributions.Normal(mean, jnp.sqrt(var))
    return jnp.expand_dims(
        ((eta - mean) * normal.cdf(eta) + var * normal.prob(eta)), -1
    )



# ---
# jupyter:
#   jupytext:
#     cell_metadata_filter: -all
#     custom_cell_magics: kql
#     text_representation:
#       extension: .py
#       format_name: percent
#       format_version: '1.3'
#       jupytext_version: 1.16.4
#   kernelspec:
#     display_name: gpjax
#     language: python
#     name: python3
# ---

# %% [markdown]
# # New to Gaussian Processes?
#
# Fantastic that you're here! This notebook is designed to be a gentle
# introduction to the mathematics of Gaussian processes (GPs). No prior
# knowledge of Bayesian inference or GPs is assumed, and this notebook is
# self-contained. At a high level, we begin by introducing Bayes' theorem and
# its implications within probabilistic modelling. We then proceed to introduce
# the Gaussian random variable along with its multivariate form. We conclude by
# showing how this notion can be extended to GPs.
#
# ## Bayes' Theorem
#
#
# A probabilistic modelling task is comprised of an observed dataset
# $\mathbf{y}$ for which we construct a model. The parameters $\theta$ of our
# model are unknown, and our goal is to conduct inference to determine their
# range of likely values. To achieve this, we apply Bayes' theorem
#
# \begin{align}
#     \label{eq:BayesTheorem}
#     p(\theta\,|\, \mathbf{y}) = \frac{p(\theta)p(\mathbf{y}\,|\,\theta)}{p(\mathbf{y})} = \frac{p(\theta)p(\mathbf{y}\,|\,\theta)}{\int_{\theta}p(\mathbf{y}, \theta)\mathrm{d}\theta}\,,
# \end{align}
#
# where $p(\mathbf{y}\,|\,\theta)$ denotes the _likelihood_, or model, and
# quantifies how likely the observed dataset $\mathbf{y}$ is, given the
# parameter estimate $\theta$. The _prior_ distribution $p(\theta)$ reflects our
# initial beliefs about the value of $\theta$ before observing data, whilst the
# _posterior_ $p(\theta\,|\, \mathbf{y})$ gives an updated estimate of the
# parameters' value, after observing $\mathbf{y}$. The _marginal likelihood_, or
# Bayesian model evidence, $p(\mathbf{y})$ is the probability of the observed
# data under all possible hypotheses that our prior model can generate. Within
# Bayesian model selection, this property makes the marginal log-likelihood an
# indispensable tool. Selecting models under this criterion places a higher
# emphasis on models that can generalise better to new data points.
#
# When the posterior distribution belongs to the same family of probability
# distributions as the prior, we describe the prior and the likelihood as
# _conjugate_ to each other. Such a scenario is convenient in Bayesian inference
# as it allows us to derive closed-form expressions for the posterior
# distribution. When the likelihood function is a member of the exponential
# family, then there exists a conjugate prior. However, the conjugate prior may
# not have a form that precisely reflects the practitioner's belief surrounding
# the parameter. For this reason, conjugate models seldom appear; one exception
# to this is GP regression that we present fully in our [Regression notebook](https://docs.jaxgaussianprocesses.com/examples/regression/).
#
# For models that do not contain a conjugate prior, the marginal log-likelihood
# must be calculated to normalise the posterior distribution and ensure it
# integrates to 1. For models with a single, 1-dimensional parameter, it may be
# possible to compute this integral analytically or through a quadrature scheme,
# such as Gauss-Hermite. However, in machine learning, the dimensionality of
# $\theta$ is often large and the corresponding integral required to compute
# $p(\mathbf{y})$ quickly becomes intractable as the dimension grows. Techniques
# such as Markov Chain Monte Carlo and variational inference allow us to
# approximate integrals such as the one seen in $p(\mathbf{y})$.
#
# Once a posterior distribution has been obtained, we can make predictions at
# new points $\mathbf{y}^{\star}$ through the _posterior predictive
# distribution_. This is achieved by integrating out the parameter set $\theta$
# from our posterior distribution through
# \begin{align}
#     p(\mathbf{y}^{\star}\mid \mathbf{y}) = \int p(\mathbf{y}^{\star} \,|\, \theta, \mathbf{y} ) p(\theta\,|\, \mathbf{y})\mathrm{d}\theta\,.
# \end{align}
# As with the marginal log-likelihood, evaluating this quantity requires
# computing an integral which may not be tractable, particularly when $\theta$
# is high-dimensional.
#
# It is difficult to communicate statistics directly through a posterior
# distribution, so we often compute and report moments of the posterior
# distribution. Most commonly, we report the first moment and the centred second
# moment
# $$
# \begin{alignat}{2}
#     \mu  = \mathbb{E}[\theta\,|\,\mathbf{y}]  & = \int \theta p(\theta\mid\mathbf{y})\mathrm{d}\theta\\
#     \sigma^2  = \mathbb{V}[\theta\,|\,\mathbf{y}] & = \int \left(\theta -
#     \mathbb{E}[\theta\,|\,\mathbf{y}]\right)^2p(\theta\,|\,\mathbf{y})\mathrm{d}\theta&\,.
# \end{alignat}
# $$
# Through this pair of statistics, we can communicate our beliefs about the most
# likely value of $\theta$ i.e., $\mu$, and the uncertainty $\sigma$ around the
# expected value. However, as with the marginal log-likelihood and predictive
# posterior distribution, computing these statistics again requires a
# potentially intractable integral.
#
# ## Gaussian random variables
#
# We begin our review with the simplest case; a univariate Gaussian random
# variable. For a random variable $y$, let $\mu\in\mathbb{R}$ be a mean scalar
# and $\sigma^2\in\mathbb{R}_{>0}$ a variance scalar. If $y$ is a Gaussian
# random variable, then the density of $y$ is
# $$
# \begin{align}
#     \mathcal{N}(y\,|\, \mu, \sigma^2) = \frac{1}{\sqrt{2\pi\sigma^2}}\exp\left(-\frac{(y-\mu)^2}{2\sigma^{2}}\right)\,.
# \end{align}
# $$
# We can plot three different parameterisations of this density.

# %%
import warnings

import jax.numpy as jnp
import jax.random as jr
import matplotlib as mpl
import matplotlib.pyplot as plt
import pandas as pd
import seaborn as sns
import tensorflow_probability.substrates.jax as tfp

from examples.utils import (
    confidence_ellipse,
    use_mpl_style,
)

# set the default style for plotting
use_mpl_style()

key = jr.key(42)


cols = mpl.rcParams["axes.prop_cycle"].by_key()["color"]
tfd = tfp.distributions

ud1 = tfd.Normal(0.0, 1.0)
ud2 = tfd.Normal(-1.0, 0.5)
ud3 = tfd.Normal(0.25, 1.5)

xs = jnp.linspace(-5.0, 5.0, 500)

fig, ax = plt.subplots()
for d in [ud1, ud2, ud3]:
    ax.plot(
        xs,
        jnp.exp(d.log_prob(xs)),
        label=f"$\\mathcal{{N}}({{{float(d.mean())}}},\\  {{{float(d.stddev())}}}^2)$",
    )
    ax.fill_between(xs, jnp.zeros_like(xs), jnp.exp(d.log_prob(xs)), alpha=0.2)
ax.legend(loc="best")

# %% [markdown]
# A Gaussian random variable is uniquely defined in distribution by its mean
# $\mu$ and variance $\sigma^2$ and we therefore write $y\sim\mathcal{N}(\mu,
# \sigma^2)$ when describing a Gaussian random variable. We can compute
# these two quantities by
# $$
# \begin{align}
#     \mathbb{E}[y] = \mu\,, \quad \quad \mathbb{E}\left[(y-\mu)^2\right] =\sigma^2\,.
# \end{align}
# $$
# Extending this concept to vector-valued random variables reveals the
# multivariate Gaussian random variables which brings us closer to the full
# definition of a GP.
#
# Let $\mathbf{y}$ be a $D$-dimensional random variable, $\boldsymbol{\mu}$ be a
# $D$-dimensional mean vector and $\boldsymbol{\Sigma}$ be a $D\times D$
# covariance matrix. If $\mathbf{y}$ is a Gaussian random variable, then the
# density of $\mathbf{y}$ is
# $$
# \begin{align}
#     \mathcal{N}(\mathbf{y}\,|\, \boldsymbol{\mu}, \boldsymbol{\Sigma}) = \frac{1}{\sqrt{2\pi}^{D/2} \lvert\boldsymbol{\Sigma}\rvert^{1/2}} \exp\left(-\frac{1}{2} \left(\mathbf{y} - \boldsymbol{\mu}\right)^T \boldsymbol{\Sigma}^{-1} \left(\mathbf{y}-\boldsymbol{\mu}\right) \right) \,.
# \end{align}
# $$
# Three example parameterisations of this can be visualised below where $\rho$
# determines the correlation of the multivariate Gaussian.

# %%
key = jr.key(123)

d1 = tfd.MultivariateNormalDiag(loc=jnp.zeros(2), scale_diag=jnp.ones(2))
d2 = tfd.MultivariateNormalTriL(
    jnp.zeros(2), jnp.linalg.cholesky(jnp.array([[1.0, 0.9], [0.9, 1.0]]))
)
d3 = tfd.MultivariateNormalTriL(
    jnp.zeros(2), jnp.linalg.cholesky(jnp.array([[1.0, -0.5], [-0.5, 1.0]]))
)

dists = [d1, d2, d3]

xvals = jnp.linspace(-5.0, 5.0, 500)
yvals = jnp.linspace(-5.0, 5.0, 500)

xx, yy = jnp.meshgrid(xvals, yvals)

pos = jnp.empty(xx.shape + (2,))
pos.at[:, :, 0].set(xx)
pos.at[:, :, 1].set(yy)

fig, (ax0, ax1, ax2) = plt.subplots(figsize=(10, 3), ncols=3, tight_layout=True)
titles = [r"$\rho = 0$", r"$\rho = 0.9$", r"$\rho = -0.5$"]

cmap = mpl.colors.LinearSegmentedColormap.from_list("custom", ["white", cols[1]], N=256)

for a, t, d in zip([ax0, ax1, ax2], titles, dists):
    d_prob = d.prob(jnp.hstack([xx.reshape(-1, 1), yy.reshape(-1, 1)])).reshape(
        xx.shape
    )
    cntf = a.contourf(xx, yy, jnp.exp(d_prob), levels=20, antialiased=True, cmap=cmap)
    for c in cntf.collections:
        c.set_edgecolor("face")
    a.set_xlim(-2.75, 2.75)
    a.set_ylim(-2.75, 2.75)
    samples = d.sample(seed=key, sample_shape=(5000,))
    xsample, ysample = samples[:, 0], samples[:, 1]
    confidence_ellipse(
        xsample, ysample, a, edgecolor="#3f3f3f", n_std=1.0, linestyle="--", alpha=0.8
    )
    confidence_ellipse(
        xsample, ysample, a, edgecolor="#3f3f3f", n_std=2.0, linestyle="--"
    )
    a.plot(0, 0, "x", color=cols[0], markersize=8, mew=2)
    a.set(xlabel="x", ylabel="y", title=t)

# %% [markdown]
# Extending the intuition given for the moments of a univariate Gaussian random
# variables, we can obtain the mean and covariance by
# $$
# \begin{align}
#    \mathbb{E}[\mathbf{y}] = \mathbf{\mu}, \quad \operatorname{Cov}(\mathbf{y}) & = \mathbf{E}\left[(\mathbf{y} - \mathbf{\mu)}(\mathbf{y} - \mathbf{\mu)}^{\top} \right]\\
#       & =\mathbb{E}[\mathbf{y}\mathbf{y}^{\top}] - \mathbb{E}[\mathbf{y}]\mathbb{E}[\mathbf{y}]^{\top} \\
#       & =\mathbf{\Sigma}\,.
# \end{align}
# $$
# The covariance matrix is a symmetric positive definite matrix that generalises
# the notion of variance to multiple dimensions. The matrix's diagonal entries
# contain the variance of each element, whilst the off-diagonal entries quantify
# the degree to which the respective pair of random variables are linearly
# related; this quantity is called the _covariance_.
#
# Assuming a Gaussian likelihood function in a Bayesian model is attractive as
# the mean and variance parameters are highly interpretable. This makes prior
# elicitation straightforward as the parameters' value can be intuitively
# contextualised within the scope of the problem at hand. Further, in models
# where the posterior distribution is Gaussian, we again use the distribution's
# mean and variance to describe our prediction and corresponding uncertainty
# around a given event occurring.
#
# Not only are Gaussian random variables highly interpretable, but linear
# operations involving them lead to analytical solutions. An example of this
# that will be useful in the sequel is the marginalisation and conditioning
# property of sets of Gaussian random variables. We will present these two
# results now for a pair of Gaussian random variables, but it should be stressed
# that these results hold for any finite set of Gaussian random variables.
#
# For a pair of random variables $\mathbf{x}$ and $\mathbf{y}$ defined on the
# same support, the distribution over them both is known as the _joint
# distribution_. The joint distribution $p(\mathbf{x}, \mathbf{y})$ quantifies
# the probability of two events, one from $p(\mathbf{x})$ and another from
# $p(\mathbf{y})$, occurring at the same time. We visualise this idea below.

# %%
n = 1000
x = tfd.Normal(loc=0.0, scale=1.0).sample(seed=key, sample_shape=(n,))
key, subkey = jr.split(key)
y = tfd.Normal(loc=0.25, scale=0.5).sample(seed=subkey, sample_shape=(n,))
key, subkey = jr.split(subkey)
xfull = tfd.Normal(loc=0.0, scale=1.0).sample(seed=subkey, sample_shape=(n * 10,))
key, subkey = jr.split(subkey)
yfull = tfd.Normal(loc=0.25, scale=0.5).sample(seed=subkey, sample_shape=(n * 10,))
key, subkey = jr.split(subkey)
df = pd.DataFrame({"x": x, "y": y, "idx": jnp.ones(n)})

with warnings.catch_warnings():
    warnings.simplefilter("ignore")
    g = sns.jointplot(
        data=df,
        x="x",
        y="y",
        hue="idx",
        marker=".",
        space=0.0,
        xlim=(-4.0, 4.0),
        ylim=(-4.0, 4.0),
        height=4,
        marginal_ticks=False,
        legend=False,
        palette="inferno",
        marginal_kws={
            "fill": True,
            "linewidth": 1,
            "color": cols[1],
            "alpha": 0.3,
            "bw_adjust": 2,
            "cmap": cmap,
        },
        joint_kws={"color": cols[1], "size": 3.5, "alpha": 0.4, "cmap": cmap},
    )
    g.ax_joint.annotate(text=r"$p(\mathbf{x}, \mathbf{y})$", xy=(-3, -1.75))
    g.ax_marg_x.annotate(text=r"$p(\mathbf{x})$", xy=(-2.0, 0.225))
    g.ax_marg_y.annotate(text=r"$p(\mathbf{y})$", xy=(0.4, -0.78))
    confidence_ellipse(
        xfull,
        yfull,
        g.ax_joint,
        edgecolor="#3f3f3f",
        n_std=1.0,
        linestyle="--",
        linewidth=0.5,
    )
    confidence_ellipse(
        xfull,
        yfull,
        g.ax_joint,
        edgecolor="#3f3f3f",
        n_std=2.0,
        linestyle="--",
        linewidth=0.5,
    )
    confidence_ellipse(
        xfull,
        yfull,
        g.ax_joint,
        edgecolor="#3f3f3f",
        n_std=3.0,
        linestyle="--",
        linewidth=0.5,
    )

# %% [markdown]
# Formally, we can define this by letting $p(\mathbf{x}, \mathbf{y})$ be the
# joint probability distribution defined over
# $\mathbf{x}\sim\mathcal{N}(\boldsymbol{\mu}_{\mathbf{x}}, \boldsymbol{\Sigma}_{\mathbf{xx}})$ and
# $\mathbf{y}\sim\mathcal{N}(\boldsymbol{\mu}_{\mathbf{y}}, \boldsymbol{\Sigma}_{\mathbf{yy}})$.
# We define the joint distribution as
# $$
# \begin{align}
#     p\left(\begin{bmatrix}
#         \mathbf{x} \\ \mathbf{y}
#     \end{bmatrix}\right) = \mathcal{N}\left(\begin{bmatrix}
#         \boldsymbol{\mu}_{\mathbf{x}} \\ \boldsymbol{\mu}_{\mathbf{y}}
#     \end{bmatrix}, \begin{bmatrix}
#         \boldsymbol{\Sigma}_{\mathbf{xx}} & \boldsymbol{\Sigma}_{\mathbf{xy}}\\
#         \boldsymbol{\Sigma}_{\mathbf{yx}} & \boldsymbol{\Sigma}_{\mathbf{yy}}
#     \end{bmatrix} \right)\,,
# \end{align}
# $$
# where $\boldsymbol{\Sigma}_{\mathbf{x}\mathbf{y}}$ is the cross-covariance
# matrix of $\mathbf{x}$ and $\mathbf{y}$.
#
# When presented with a joint distribution, two tasks that we may wish to
# perform are _marginalisation_ and _conditioning_. For a joint distribution
# $p(\mathbf{x}, \mathbf{y})$ where we are interested only in $p(\mathbf{x})$,
# we must integrate over all possible values of $\mathbf{y}$ to obtain
# $p(\mathbf{x})$. This process is marginalisation. Conditioning allows us to
# evaluate the probability of one random variable, given that the other random
# variable is fixed. For a joint Gaussian distribution, marginalisation and
# conditioning have analytical expressions where the resulting distribution is
# also a Gaussian random variable.
#
# For a joint Gaussian random variable, the marginalisation of $\mathbf{x}$ or
# $\mathbf{y}$ is given by
# $$
# \begin{alignat}{3}
#     & \int p(\mathbf{x}, \mathbf{y})\mathrm{d}\mathbf{y} && = p(\mathbf{x})
#     && = \mathcal{N}(\boldsymbol{\mu}_{\mathbf{x}},\boldsymbol{\Sigma}_{\mathbf{xx}})\\
#     & \int p(\mathbf{x}, \mathbf{y})\mathrm{d}\mathbf{x} && = p(\mathbf{y})
#     && = \mathcal{N}(\boldsymbol{\mu}_{\mathbf{y}},
#     \boldsymbol{\Sigma}_{\mathbf{yy}})\,.
# \end{alignat}
# $$
# The conditional distributions are given by
# $$
# \begin{align}
#     p(\mathbf{y}\,|\, \mathbf{x}) & = \mathcal{N}\left(\boldsymbol{\mu}_{\mathbf{y}} + \boldsymbol{\Sigma}_{\mathbf{yx}}\boldsymbol{\Sigma}_{\mathbf{xx}}^{-1}(\mathbf{x}-\boldsymbol{\mu}_{\mathbf{x}}), \boldsymbol{\Sigma}_{\mathbf{yy}}-\boldsymbol{\Sigma}_{\mathbf{yx}}\boldsymbol{\Sigma}_{\mathbf{xx}}^{-1}\boldsymbol{\Sigma}_{\mathbf{xy}}\right)\,.
# \end{align}
# $$
#
# Within this section, we have introduced the idea of multivariate Gaussian
# random variables and presented some key results concerning their properties.
# In the following section, we will lift our presentation of Gaussian random
# variables to GPs.
#
# ## Gaussian processes
#
# When transitioning from Gaussian random variables to GP there is a shift in
# thought required to parse the forthcoming material. Firstly, to be consistent
# with the general literature, we hereon use $\mathbf{X}$ to denote an observed
# vector of data points, not a random variable as has been true up until now. To
# distinguish between matrices and vectors, we use bold upper case characters
# e.g., $\mathbf{X}$ for matrices, and bold lower case characters for vectors
# e.g., $\mathbf{x}$.
#
# We are interested in modelling supervised learning problems, where we have $n$
# observations $\mathbf{y}=\{y_1, y_2,\ldots ,y_n\}\subset\mathcal{Y}$ at
# corresponding inputs
# $\mathbf{X}=\{\mathbf{x}_1,\mathbf{x}_2,\ldots,\mathbf{x}_n\}\subset\mathcal{X}$.
# We aim to capture the relationship between $\mathbf{X}$ and $\mathbf{y}$ using
# a model $f$ with which we may make predictions at an unseen set of test points
# $\mathbf{X}^{\star}\subset\mathcal{X}$. We formalise this by
# $$
# \begin{align}
#     y = f(\mathbf{X}) + \varepsilon\,,
# \end{align}
# $$
# where $\varepsilon$ is an observational noise term. We collectively refer to
# $(\mathbf{X}, \mathbf{y})$ as the _training data_ and $\mathbf{X}^{\star}$ as
# the set of _test points_.
# This process is visualised below
#
# ![](intro_to_gps/generating_process.png)
#
# As we shall go on to see, GPs offer an appealing workflow for scenarios such
# as this, all under a Bayesian framework.
#
# We write a GP $f(\cdot) \sim \mathcal{GP}(\mu(\cdot), k(\cdot, \cdot))$ with
# mean function $\mu: \mathcal{X} \rightarrow \mathbb{R}$ and
# $\boldsymbol{\theta}$-parameterised kernel
# $k: \mathcal{X} \times \mathcal{X}\rightarrow \mathbb{R}$.
# When evaluating the GP on a finite set of points
# $\mathbf{X}\subset\mathcal{X}$, $k$ gives rise to the Gram matrix
# $\mathbf{K}_{ff}$ such that the $(i, j)^{\text{th}}$ entry of the matrix is
# given by $[\mathbf{K}_{ff}]_{i, j} = k(\mathbf{x}_i, \mathbf{x}_j)$. As is
# conventional within the literature, we centre our training data and assume
# $\mu(\mathbf{X}):= 0$ for all $\mathbf{X}\in\mathcal{X}$. We further drop
# dependency on $\boldsymbol{\theta}$ and $\mathbf{X}$ for notational
# convenience in the remainder of this article.
#
# We define a joint GP prior over the latent function
# $$
# \begin{align}
#     p(\mathbf{f}, \mathbf{f}^{\star}) = \mathcal{N}\left(\mathbf{0}, \begin{bmatrix}
#         \mathbf{K}_{xf} & \mathbf{K}_{xx}
#     \end{bmatrix}\right)\,,
# \end{align}
# $$
# where $\mathbf{f}^{\star} = f(\mathbf{X}^{\star})$. Conditional on the GP's
# latent function $f$, we assume a factorising likelihood generates our
# observations
# $$
# \begin{align}
#     p(\mathbf{y}\,|\,\mathbf{f}) = \prod_{i=1}^n p(y_i\,|\, f_i)\,.
# \end{align}
# $$
# Strictly speaking, the likelihood function is
# $p(\mathbf{y}\,|\,\phi(\mathbf{f}))$ where $\phi$ is the likelihood function's
# associated link function. Example link functions include the probit or
# logistic functions for a Bernoulli likelihood and the identity function for a
# Gaussian likelihood. We eschew this notation for now as this section primarily
# considers Gaussian likelihood functions where the role of $\phi$ is
# superfluous. However, this intuition will be helpful for models with a
# non-Gaussian likelihood, such as those encountered in
# [classification](https://docs.jaxgaussianprocesses.com/examples/classification).
#
# Applying Bayes' theorem \eqref{eq:BayesTheorem} yields the joint posterior distribution over the
# latent function
# $$
# \begin{align}
#     p(\mathbf{f}, \mathbf{f}^{\star}\,|\,\mathbf{y}) = \frac{p(\mathbf{y}\,|\,\mathbf{f})p(\mathbf{f},\mathbf{f}^{\star})}{p(\mathbf{y})}\,.
# \end{align}
# $$
#
# The choice of kernel function that we use to parameterise our GP is an
# important modelling decision as the choice of kernel dictates properties such
# as differentiability, variance and characteristic lengthscale of the functions
# that are admissible under the GP prior. A kernel is a positive-definite
# function with parameters $\boldsymbol{\theta}$ that maps pairs of inputs
# $\mathbf{X}, \mathbf{X}' \in \mathcal{X}$ onto the real line. We dedicate the
# entirety of the [Introduction to Kernels
# notebook](https://docs.jaxgaussianprocesses.com/examples/intro_to_kernels) to
# exploring the different GPs each kernel can yield.
#
# ## Gaussian process regression
#
# When the likelihood function is a Gaussian distribution
# $p(y_i\,|\, f_i) = \mathcal{N}(y_i\,|\, f_i, \sigma_n^2)$,
# marginalising $\mathbf{f}$ from the joint posterior to obtain
# the posterior predictive distribution is exact
# $$
# \begin{align}
#     p(\mathbf{f}^{\star}\mid \mathbf{y}) = \mathcal{N}(\mathbf{f}^{\star}\,|\,\boldsymbol{\mu}_{\,|\,\mathbf{y}}, \Sigma_{\,|\,\mathbf{y}})\,,
# \end{align}
# $$
# where
# $$
# \begin{align}
#     \mathbf{\mu}_{\mid \mathbf{y}} & = \mathbf{K}_{\star f}\left( \mathbf{K}_{ff}+\sigma^2_n\mathbf{I}_n\right)^{-1}\mathbf{y} \\
#     \Sigma_{\,|\,\mathbf{y}} & = \mathbf{K}_{\star\star} - \mathbf{K}_{xf}\left(\mathbf{K}_{ff} + \sigma_n^2\mathbf{I}_n\right)^{-1}\mathbf{K}_{fx} \,.
# \end{align}
# $$
# Further, the log of the  marginal likelihood of the GP can
# be analytically expressed as
# $$
# \begin{align}
#         & = 0.5\left(-\underbrace{\mathbf{y}^{\top}\left(\mathbf{K}_{ff} + \sigma_n^2\mathbf{I}_n \right)^{-1}\mathbf{y}}_{\text{Data fit}} -\underbrace{\log\lvert \mathbf{K}_{ff} + \sigma^2_n\rvert}_{\text{Complexity}} -\underbrace{n\log 2\pi}_{\text{Constant}} \right)\,.
# \end{align}
# $$
#
# Model selection can be performed for a GP through gradient-based optimisation
# of $\log p(\mathbf{y})$ with respect to the kernel's parameters
# $\boldsymbol{\theta}$ and the observational noise $\sigma^2_n$. Collectively,
# we call these terms the model hyperparameters
# $\boldsymbol{\xi} = \{\boldsymbol{\theta},\sigma_n^2\}$
# from which the maximum likelihood estimate is given by
# $$
# \begin{align*}
#     \boldsymbol{\xi}^{\star} = \operatorname{argmax}_{\boldsymbol{\xi} \in \Xi} \log p(\mathbf{y})\,.
# \end{align*}
# $$
#
#
# Observing the individual terms in the marginal log-likelihood can help
# understand exactly why optimising the marginal log-likelihood gives reasonable
# solutions. The _data fit_ term is the only component of the marginal
# log-likelihood that includes the observed response $\mathbf{y}$ and will
# therefore encourage solutions that model the data well. Conversely, the
# _complexity_ term contains a determinant operator and therefore measures the
# _volume_ of the function space covered by the GP. Whilst a more complex
# function has a better chance of modelling the observed data well, this is only
# true to a point and functions that are overly complex will overfit the data.
# Optimising with respect to the marginal log-likelihood balances these two
# objectives when identifying the optimal solution, as visualised below.
#
# ![](intro_to_gps/decomposed_mll.png)
#
# ## Conclusions
#
# Within this notebook we have built up the concept of a GP, starting from
# Bayes' theorem and the definition of a Gaussian random variable. Using the
# ideas presented in this notebook, the user should be in a position to dive
# into our [Regression
# notebook](https://docs.jaxgaussianprocesses.com/examples/regression/) and
# start getting their hands on some code. For those looking to learn more about
# the underling theory of GPs, an excellent starting point is the [Gaussian
# Processes for Machine Learning](http://gaussianprocess.org/gpml/) textbook.
# Alternatively, the [thesis of Alexander
# Terenin](https://arxiv.org/abs/2202.10613) provides a rigorous exposition of
# GPs that served as the inspiration for this notebook.
#
# ## System Configuration

# %%
# %reload_ext watermark
# %watermark -n -u -v -iv -w -a 'Thomas Pinder'



# Copyright 2023 The GPJax Contributors. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ==============================================================================
from jax import config

config.update("jax_enable_x64", True)

import jax.numpy as jnp
import jax.random as jr
import pytest

from gpjax.decision_making.test_functions.continuous_functions import (
    AbstractContinuousTestFunction,
    Forrester,
    LogarithmicGoldsteinPrice,
    Quadratic,
)
from gpjax.decision_making.utility_maximizer import (
    AbstractSinglePointUtilityMaximizer,
    ContinuousSinglePointUtilityMaximizer,
    _get_discrete_maximizer,
)
from gpjax.typing import KeyArray


def test_abstract_single_batch_utility_maximizer():
    with pytest.raises(TypeError):
        AbstractSinglePointUtilityMaximizer()


@pytest.mark.parametrize(
    "test_function, dimensionality",
    [(Forrester(), 1), (LogarithmicGoldsteinPrice(), 2)],
)
@pytest.mark.parametrize("key", [jr.key(42), jr.key(10)])
def test_discrete_maximizer_returns_correct_point(
    test_function: AbstractContinuousTestFunction,
    dimensionality: int,
    key: KeyArray,
):
    query_points = test_function.generate_test_points(1000, key=key)
    utility_function = lambda x: -1.0 * test_function.evaluate(x)
    utility_vals = utility_function(query_points)
    true_max_utility_val = jnp.max(utility_vals)
    discrete_maximizer = _get_discrete_maximizer(query_points, utility_function)
    assert discrete_maximizer.shape == (1, dimensionality)
    assert discrete_maximizer.dtype == jnp.float64
    assert utility_function(discrete_maximizer)[0][0] == true_max_utility_val


@pytest.mark.parametrize("num_initial_samples", [0, -1, -10])
def test_continuous_maximizer_raises_error_with_erroneous_num_initial_samples(
    num_initial_samples: int,
):
    with pytest.raises(ValueError):
        ContinuousSinglePointUtilityMaximizer(
            num_initial_samples=num_initial_samples, num_restarts=1
        )


@pytest.mark.parametrize("num_restarts", [0, -1, -10])
def test_continuous_maximizer_raises_error_with_erroneous_num_restarts(
    num_restarts: int,
):
    with pytest.raises(ValueError):
        ContinuousSinglePointUtilityMaximizer(
            num_initial_samples=1, num_restarts=num_restarts
        )


@pytest.mark.parametrize(
    "test_function, dimensionality",
    [(Forrester(), 1), (LogarithmicGoldsteinPrice(), 2)],
)
@pytest.mark.parametrize("key", [jr.key(42), jr.key(10)])
@pytest.mark.parametrize("num_restarts", [1, 3])
def test_continous_maximizer_returns_same_point_with_same_key(
    test_function: AbstractContinuousTestFunction,
    dimensionality: int,
    key: KeyArray,
    num_restarts: int,
):
    continuous_maximizer_one = ContinuousSinglePointUtilityMaximizer(
        num_initial_samples=1000, num_restarts=num_restarts
    )
    continuous_maximizer_two = ContinuousSinglePointUtilityMaximizer(
        num_initial_samples=1000, num_restarts=num_restarts
    )
    utility_function = lambda x: -1.0 * test_function.evaluate(x)
    maximizer_one = continuous_maximizer_one.maximize(
        utility_function=utility_function,
        search_space=test_function.search_space,
        key=key,
    )
    maximizer_two = continuous_maximizer_two.maximize(
        utility_function=utility_function,
        search_space=test_function.search_space,
        key=key,
    )
    assert maximizer_one.shape == (1, dimensionality)
    assert maximizer_one.dtype == jnp.float64
    assert maximizer_two.shape == (1, dimensionality)
    assert maximizer_two.dtype == jnp.float64
    assert jnp.equal(maximizer_one, maximizer_two).all()


@pytest.mark.parametrize(
    "test_function, dimensionality",
    [
        (Forrester(), 1),
        (LogarithmicGoldsteinPrice(), 2),
    ],
)
@pytest.mark.parametrize("key", [jr.key(42), jr.key(10)])
@pytest.mark.parametrize("num_restarts", [1, 3])
def test_continuous_maximizer_finds_correct_point(
    test_function: AbstractContinuousTestFunction,
    dimensionality: int,
    key: KeyArray,
    num_restarts: int,
):
    continuous_utility_maximizer = ContinuousSinglePointUtilityMaximizer(
        num_initial_samples=1000, num_restarts=num_restarts
    )
    utility_function = lambda x: -1.0 * test_function.evaluate(x)
    true_utility_maximizer = test_function.minimizer
    maximizer = continuous_utility_maximizer.maximize(
        utility_function=utility_function,
        search_space=test_function.search_space,
        key=key,
    )
    assert maximizer.shape == (1, dimensionality)
    assert maximizer.dtype == jnp.float64
    assert jnp.allclose(maximizer, true_utility_maximizer, atol=1e-6).all()


@pytest.mark.parametrize("key", [jr.key(42), jr.key(10), jr.key(1)])
@pytest.mark.parametrize("num_restarts", [1, 3])
def test_continuous_maximizer_jaxopt_component(key: KeyArray, num_restarts: int):
    quadratic = Quadratic()
    continuous_utility_maximizer = ContinuousSinglePointUtilityMaximizer(
        num_initial_samples=1,  # Force JaxOpt L-GFBS-B to do the heavy lifting
        num_restarts=num_restarts,
    )
    utility_function = lambda x: -1.0 * quadratic.evaluate(x)
    true_utility_maximizer = quadratic.minimizer
    maximizer = continuous_utility_maximizer.maximize(
        utility_function=utility_function,
        search_space=quadratic.search_space,
        key=key,
    )
    assert maximizer.shape == (1, 1)
    assert maximizer.dtype == jnp.float64
    assert jnp.allclose(maximizer, true_utility_maximizer, atol=1e-6).all()
